from collections.abc import Callable, Iterable
from enum import Enum, auto
from functools import wraps
from typing import Any, ParamSpec, Protocol, TypeVar

import numpy as np
import sympy

from ..._types import FloatNDArray

T = TypeVar('T')
P = ParamSpec('P')
R = TypeVar('R')


class Direction(Enum):
    """Optimization direction of metric."""

    MAXIMIZE = auto()
    MINIMIZE = auto()


class MetricFn(Protocol):  # noqa: D101
    def __call__(self, y_true: FloatNDArray, y_score: FloatNDArray, **kwargs: Any) -> float: ...  # noqa: D102


class LogitConsts(Protocol):  # noqa: D101
    def prepare(  # noqa: D102
        self, x: FloatNDArray, y_true: FloatNDArray, **kwargs: Any
    ) -> tuple[FloatNDArray, FloatNDArray, FloatNDArray]: ...


class BoostGradientConst(Protocol):  # noqa: D101
    def __call__(self, y_true: FloatNDArray, **kwargs: Any) -> FloatNDArray: ...  # noqa: D102


class ThresholdFn(Protocol):  # noqa: D101
    def __call__(  # noqa: D102
        self, y_true: FloatNDArray, y_score: FloatNDArray, **kwargs: Any
    ) -> FloatNDArray | float: ...


class RateFn(Protocol):  # noqa: D101
    def __call__(  # noqa: D102
        self, y_true: FloatNDArray, y_score: FloatNDArray, **kwargs: Any
    ) -> float: ...


class SympyFnPickleMixin:
    """
    Class to allow pickling of classes which use lambdified functions.

    Lambdified are deleted before pickling and restored after unpickling.
    Define a class attribute in subclasses: _sympy_functions = {'func_name': 'equation_attr_name'}.
    """

    def __getstate__(self) -> dict[str, Any]:
        state = self.__dict__.copy()
        for func_name in getattr(self, '_sympy_functions', {}):
            if func_name in state:
                del state[func_name]
        return state

    def __setstate__(self, state: dict[str, Any]) -> None:
        self.__dict__.update(state)
        for func_name, eqn_attr in getattr(self, '_sympy_functions', {}).items():
            equation = getattr(self, eqn_attr)
            if sympy.simplify(equation).is_Number and not equation.free_symbols:
                setattr(self, func_name, lambda *args, **kwargs: float(equation))  # noqa: B023
            else:
                setattr(self, func_name, sympy.lambdify(list(equation.free_symbols), equation))


def _check_parameters_decorator(*parameters: str | sympy.Expr) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """
    Check if all parameters are provided.

    In particular:
        - deterministic parameters
        - distribution parameters of stochastic variables
    """

    def decorator(func: Callable[P, R]) -> Callable[P, R]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            for value in parameters:
                if str(value) not in kwargs:
                    raise ValueError(f'Metric expected a value for {value}, did not receive it.')
            return func(*args, **kwargs)

        return wrapper

    return decorator


def _check_parameters(symbols: Iterable[str | sympy.Expr], kwargs: dict[str, Any]) -> None:
    """
    Check if all parameters are provided.

    In particular:
        - deterministic parameters
        - distribution parameters of stochastic variables
    """
    for symbol in symbols:
        if str(symbol) not in kwargs:
            raise ValueError(f'Metric expected a value for {symbol}, did not receive it.')


def _filter_parameters(
    expression: sympy.Expr, parameters: dict[str, float | FloatNDArray]
) -> dict[str, float | FloatNDArray]:
    """
    Filter the parameters dictionary to only include those that are free symbols in the expression.

    Parameters
    ----------
    expression : sympy.Expr
        The expression to filter the parameters against.

    parameters : dict[str, float | FloatNDArray]
        The parameters dictionary to filter.
        Keys are parameter names and values are their corresponding values.

    Returns
    -------
    filtered_parameters : dict[str, float | FloatNDArray]
        A dictionary containing only the parameters that are free symbols in the expression.
    """
    free_symbols = {str(symbol) for symbol in expression.free_symbols}
    filtered_parameters = {key: value for key, value in parameters.items() if key in free_symbols}
    return filtered_parameters


def _evaluate_expression(expression: sympy.Expr, **parameters: FloatNDArray | float) -> FloatNDArray | float:
    """
    Evaluate a sympy expression with the given parameters.

    Parameters
    ----------
    expression : sympy.Expr
        The sympy expression to convert.
    parameters : float or NDArray of shape (n_samples,)
        The parameter values for the costs and benefits defined in the metric.
        If any parameter is a stochastic variable, you should pass values for their distribution parameters.
        You can set the parameter values for either the symbol names or their aliases.

    Returns
    -------
    function : callable
        A numpy function that computes the value of the expression with the given parameters.
    """
    filtered_parameters = _filter_parameters(expression, parameters)
    result: float | FloatNDArray = sympy.lambdify(list(expression.free_symbols), expression)(**filtered_parameters)
    return result


def _safe_lambdify(
    expression: sympy.Expr, variables: Iterable[sympy.Symbol] | None = None
) -> Callable[..., FloatNDArray | float]:
    """
    Safely lambdify a sympy expression.

    If the expression is a constant, return a function that returns the constant value.

    Parameters
    ----------
    expression : sympy.Expr
        The sympy expression to convert.
    variables : Iterable[sympy.Symbol], optional
        The variables to use in the lambdified function.
        If None, all free symbols in the expression will be used.

    Returns
    -------
    function : callable
        A numpy function that computes the value of the expression.
    """
    if sympy.simplify(expression).is_Number and not expression.free_symbols:
        return lambda *args, **kwargs: float(expression)
    if variables is None:
        variables = list(expression.free_symbols)
    function: Callable[..., float | FloatNDArray] = sympy.lambdify(variables, expression)
    return function


def _safe_run_lambda(
    function: Callable[..., T],
    expression: sympy.Expr,
    **parameters: FloatNDArray | float,
) -> T:
    """
    Safely evaluate a lambdified expression with the given parameters.

    Parameters
    ----------
    function : callable
        A lambda function that computes the value of the expression.
    expression : sympy.Expr
        The sympy expression to convert.
    parameters : float or NDArray of shape (n_samples,)
        The parameter values for the costs and benefits defined in the metric.
        If any parameter is a stochastic variable, you should pass values for their distribution parameters.
        You can set the parameter values for either the symbol names or their aliases.

    Returns
    -------
    value : Any
        The result of evaluating the function with the given parameters.
    """
    filtered_parameters = _filter_parameters(expression, parameters)
    return function(**filtered_parameters)


def _safe_run_lambda_array(
    function: Callable[..., FloatNDArray | float],
    expression: sympy.Expr,
    shape: int | tuple[int, ...],
    **parameters: FloatNDArray | float,
) -> FloatNDArray:
    """
    Safely evaluate a lambdified expression with the given parameters and enforces it to be an array of the given shape.

    Parameters
    ----------
    function : callable
        A lambda function that computes the value of the expression.
    expression : sympy.Expr
        The sympy expression to convert.
    shape : int or tuple of int
        Shape of the output array.
    parameters : float or NDArray of shape (n_samples,)
        The parameter values for the costs and benefits defined in the metric.
        If any parameter is a stochastic variable, you should pass values for their distribution parameters.
        You can set the parameter values for either the symbol names or their aliases.

    Returns
    -------
    value : FloatNDArray
        The result of evaluating the function with the given parameters.
    """
    filtered_parameters = _filter_parameters(expression, parameters)
    output = function(**filtered_parameters)
    if isinstance(output, float):
        output = np.full(shape, output)
    return output



from dataclasses import dataclass
from os.path import dirname, join
from typing import Generic, Literal, TypeVar, overload

import numpy as np
import pandas as pd

from .._types import FloatNDArray

Frame = TypeVar('Frame', pd.DataFrame, FloatNDArray)
Series = TypeVar('Series', pd.Series, FloatNDArray)


@dataclass(frozen=True)
class Dataset(Generic[Frame, Series]):
    """
    Container object for datasets returned by the load functions.

    Attributes
    ----------
    data : :class:`pandas:pandas.DataFrame` or :class:`numpy:numpy.ndarray`
        Features of the dataset.
    target : :class:`pandas:pandas.Series` or :class:`numpy:numpy.ndarray`
        The classification labels.
    tp_cost : :class:`pandas:pandas.Series`, :class:`numpy:numpy.ndarray` or float
        The cost of true positives.
    tn_cost : :class:`pandas:pandas.Series`, :class:`numpy:numpy.ndarray` or float
        The cost of true negatives.
    fp_cost : :class:`pandas:pandas.Series`, :class:`numpy:numpy.ndarray` or float
        The cost of false positives.
    fn_cost : :class:`pandas:pandas.Series`, :class:`numpy:numpy.ndarray` or float
        The cost of false negatives.
    feature_names : :class:`pandas:pandas.Series` or :class:`numpy:numpy.ndarray`
        The meaning of the features.
    target_names : :class:`pandas:pandas.Series` or :class:`numpy:numpy.ndarray`
        The meaning of the labels.
    name : str
        The name of the dataset.
    DESCR : str
        The full description of the dataset
    """

    data: Frame
    target: Series
    tp_cost: Series | float
    tn_cost: Series | float
    fp_cost: Series | float
    fn_cost: Series | float
    feature_names: Series
    target_names: Series
    name: str
    DESCR: str


@overload
def load_churn_tv_subscriptions(
    *,
    as_frame: Literal[True],
    return_X_y_costs: Literal[False] = False,
) -> Dataset[pd.DataFrame, pd.Series]: ...


@overload
def load_churn_tv_subscriptions(
    *,
    as_frame: Literal[False] = False,
    return_X_y_costs: Literal[False] = False,
) -> Dataset[FloatNDArray, FloatNDArray]: ...


@overload
def load_churn_tv_subscriptions(
    *,
    as_frame: Literal[True],
    return_X_y_costs: Literal[True],
) -> tuple[pd.DataFrame, pd.Series, pd.Series, pd.Series, pd.Series, pd.Series]: ...


@overload
def load_churn_tv_subscriptions(
    *,
    as_frame: Literal[False] = False,
    return_X_y_costs: Literal[True],
) -> tuple[FloatNDArray, FloatNDArray, FloatNDArray, FloatNDArray, FloatNDArray, FloatNDArray]: ...


def load_churn_tv_subscriptions(
    *, as_frame: bool = False, return_X_y_costs: bool = False
) -> (
    Dataset[pd.DataFrame, pd.Series]
    | Dataset[FloatNDArray, FloatNDArray]
    | tuple[pd.DataFrame, pd.Series, pd.Series, pd.Series, pd.Series, pd.Series]
    | tuple[FloatNDArray, FloatNDArray, FloatNDArray, FloatNDArray, FloatNDArray, FloatNDArray]
):
    """
    Load the TV Subscription Churn dataset (binary classification).

    The goal is to predict whether a customer will churn or not.
    The target variable is whether the customer churned, 'yes' = 1 and 'no' = 0.

    This dataset is from a TV cable provider containing all 9410 customers active during the first semester of 2014.
    Features names are anonymized to protect the privacy of the customers.

    For additional information about the dataset, consult the :ref:`User Guide <churn_tv_subscriptions>`.

    =================   ==============
    Classes                          2
    Churners                       455
    Non-churners                  8955
    Samples                       9410
    Features                        45
    =================   ==============

    Parameters
    ----------
    as_frame : bool, default=False
        If True, the output will be a pandas DataFrames or Series instead of numpy arrays.
    return_X_y_costs : bool, default=False
        If True, return (data, target, tp_cost, fp_cost, tn_cost, fn_cost) instead of a Dataset object.

    Returns
    -------
    dataset : :class:`~empulse.datasets.Dataset` or tuple of (data, target, tp_cost, fp_cost, tn_cost, fn_cost)
        Returns a Dataset object if `return_X_y_costs=False` (default), otherwise a tuple.

    Notes
    -----
    Cost matrix

    .. list-table::

        * -
          - Actual positive :math:`y_i = 1`
          - Actual negative :math:`y_i = 0`
        * - Predicted positive :math:`\\hat{y}_i = 1`
          - ``tp_cost`` :math:`= \\gamma_i d_i + (1 - \\gamma_i) (CLV_i + c_i)`
          - ``fp_cost`` :math:`= d_i + c_i`
        * - Predicted negative :math:`\\hat{y}_i = 0`
          - ``fn_cost`` :math:`= CLV_i`
          - ``tn_cost`` :math:`= 0`

    with
        - :math:`\\gamma_i` : probability of the customer accepting the retention offer
        - :math:`CLV_i` : customer lifetime value of the retained customer
        - :math:`d_i` : cost of incentive offered to the customer
        - :math:`c_i` : cost of contacting the customer

    References
    ----------
    .. [1] A. Correa Bahnsen, D.Aouada, B, Ottersten,
           `"A novel cost-sensitive framework for customer churn predictive modeling"
           <http://www.decisionanalyticsjournal.com/content/pdf/s40165-015-0014-6.pdf>`__,
           Decision Analytics, 2:5, 2015.

    Examples
    --------

    .. code-block:: python

        from empulse.datasets import load_churn_tv_subscriptions
        from sklearn.model_selection import train_test_split

        dataset = load_churn_tv_subscriptions()
        X_train, X_test, y_train, y_test = train_test_split(
            dataset.data, dataset.target, random_state=42
        )
    """
    module_path = dirname(__file__)
    raw_data = pd.read_csv(
        join(module_path, 'data', 'churn_tv_subscriptions.csv.gz'), delimiter=',', compression='gzip'
    )
    with open(join(module_path, 'descriptions', 'churn_tv_subscriptions.rst'), encoding='utf-8') as f:
        description = f.read()
    data = raw_data.iloc[:, 1:-5]

    if return_X_y_costs:
        if as_frame:
            return data, raw_data.target, raw_data['C_TP'], raw_data['C_FP'], raw_data['C_TN'], raw_data['C_FN']
        else:
            return (
                data.to_numpy(),
                raw_data.target.to_numpy().astype(np.int8),
                raw_data['C_TP'].to_numpy(),
                raw_data['C_FP'].to_numpy(),
                raw_data['C_TN'].to_numpy(),
                raw_data['C_FN'].to_numpy(),
            )
    else:
        return Dataset(
            data=data.to_numpy() if not as_frame else data,
            target=raw_data.target.to_numpy().astype(np.int8) if not as_frame else raw_data['target'],
            tp_cost=raw_data.C_TP.to_numpy() if not as_frame else raw_data['C_TP'],
            fp_cost=raw_data.C_FP.to_numpy() if not as_frame else raw_data['C_FP'],
            tn_cost=raw_data.C_TN.to_numpy() if not as_frame else raw_data['C_TN'],
            fn_cost=raw_data.C_FN.to_numpy() if not as_frame else raw_data['C_FN'],
            feature_names=data.columns.to_numpy() if not as_frame else data.columns,
            target_names=np.array(['no churn', 'churn']) if not as_frame else pd.Series(['no churn', 'churn']),
            name='Churn TV subscriptions',
            DESCR=description,
        )


@overload
def load_upsell_bank_telemarketing(
    *,
    as_frame: Literal[True],
    return_X_y_costs: Literal[False] = False,
    interest_rate: float = 0.02463333,
    term_deposit_fraction: float = 0.25,
    contact_cost: float = 1,
) -> Dataset[pd.DataFrame, pd.Series]: ...


@overload
def load_upsell_bank_telemarketing(
    *,
    as_frame: Literal[False] = False,
    return_X_y_costs: Literal[False] = False,
    interest_rate: float = 0.02463333,
    term_deposit_fraction: float = 0.25,
    contact_cost: float = 1,
) -> Dataset[FloatNDArray, FloatNDArray]: ...


@overload
def load_upsell_bank_telemarketing(
    *,
    as_frame: Literal[True],
    return_X_y_costs: Literal[True],
    interest_rate: float = 0.02463333,
    term_deposit_fraction: float = 0.25,
    contact_cost: float = 1,
) -> tuple[pd.DataFrame, pd.Series, pd.Series, pd.Series, pd.Series, pd.Series]: ...


@overload
def load_upsell_bank_telemarketing(
    *,
    as_frame: Literal[False] = False,
    return_X_y_costs: Literal[True],
    interest_rate: float = 0.02463333,
    term_deposit_fraction: float = 0.25,
    contact_cost: float = 1,
) -> tuple[FloatNDArray, FloatNDArray, FloatNDArray, FloatNDArray, FloatNDArray, FloatNDArray]: ...


def load_upsell_bank_telemarketing(
    *,
    as_frame: bool = False,
    return_X_y_costs: bool = False,
    interest_rate: float = 0.02463333,
    term_deposit_fraction: float = 0.25,
    contact_cost: float = 1,
) -> (
    Dataset[pd.DataFrame, pd.Series]
    | Dataset[FloatNDArray, FloatNDArray]
    | tuple[pd.DataFrame, pd.Series, pd.Series, pd.Series, pd.Series, pd.Series]
    | tuple[FloatNDArray, FloatNDArray, FloatNDArray, FloatNDArray, FloatNDArray, FloatNDArray]
):
    """
    Load the bank telemarketing dataset (binary classification).

    The goal is to predict whether a client will subscribe to a term deposit after being called by the bank.
    The target variable is whether the client subscribed to the term deposit, 'yes' = 1 and 'no' = 0.

    The dataset is related to a direct marketing campaigns (phone calls) of a Portuguese banking institution.
    The marketing campaigns were based on phone calls.
    Often, more than one contact to the same client was required,
    in order to access if the product (bank term deposit) would be or not subscribed.

    Features recorded before the contact event are removed from the original dataset [1]_ to avoid data leakage.
    Only clients with a positive balance are considered, since clients in debt are not eligible for term deposits.

    For a full data description and additional information about the dataset,
    consult the :ref:`User Guide <upsell_bank_telemarketing>`.

    =================   ==============
    Classes                          2
    Subscribers                   4787
    Non-subscribers              33144
    Samples                      37931
    Features                        10
    =================   ==============

    Parameters
    ----------
    as_frame : bool, default=False
        If True, the output will be a pandas DataFrames or Series instead of numpy arrays.
    return_X_y_costs : bool, default=False
        If True, return (data, target, tp_cost, fp_cost, tn_cost, fn_cost) instead of a Dataset object.
    interest_rate : float, default=0.02463333
        Interest rate of the term deposit.
    term_deposit_fraction : float, default=0.25
        Fraction of the client's balance that is deposited in the term deposit.
    contact_cost : float, default=1
        Cost of contacting the client.

    Returns
    -------
    dataset : :class:`~empulse.datasets.Dataset` or tuple of (data, target, tp_cost, fp_cost, tn_cost, fn_cost)
        Returns a Dataset object if `return_X_y_costs=False` (default), otherwise a tuple.

    Notes
    -----
    Cost matrix

    .. list-table::

        * -
          - Actual positive :math:`y_i = 1`
          - Actual negative :math:`y_i = 0`
        * - Predicted positive :math:`\\hat{y}_i = 1`
          - ``tp_cost`` :math:`= c`
          - ``fp_cost`` :math:`= c`
        * - Predicted negative :math:`\\hat{y}_i = 0`
          - ``fn_cost`` :math:`= r \\, d_i \\, b_i`
          - ``tn_cost`` :math:`= 0`

    with
        - :math:`c` : cost of contacting the client
        - :math:`r` : interest rate of the term deposit
        - :math:`d_i` : fraction of the client's balance that is deposited in the term deposit
        - :math:`b_i` : client's balance

    Using default parameters, it is assumed that :math:`c = 1`, :math:`r = 0.02463333`, :math:`d_i = 0.25`
    for all clients.

    References
    ----------
    .. [1] Moro, S., Rita, P., & Cortez, P. (2014).
           Bank Marketing [Dataset]. UCI Machine Learning Repository. https://doi.org/10.24432/C5K306.

    .. [2] S. Moro, R. Laureano and P. Cortez.
           Using Data Mining for Bank Direct Marketing: An Application of the CRISP-DM Methodology.
           In P. Novais et al. (Eds.),
           Proceedings of the European Simulation and Modelling Conference
           - ESM'2011, pp. 117-121, Guimaraes, Portugal, October, 2011. EUROSIS. [bank.zip]

    .. [3] A. Correa Bahnsen, A. Stojanovic, D.Aouada, B, Ottersten,
           `"Improving Credit Card Fraud Detection with Calibrated Probabilities"
           <http://albahnsen.com/files/%20Improving%20Credit%20Card%20Fraud%20Detection%20by%20using%20Calibrated%20Probabilities%20-%20Publish.pdf>`__,
           in Proceedings of the fourteenth SIAM International Conference on Data Mining,
           677-685, 2014.

    Examples
    --------

    .. code-block:: python

        from empulse.datasets import load_upsell_bank_telemarketing
        from sklearn.model_selection import train_test_split

        dataset = load_upsell_bank_telemarketing()
        X_train, X_test, y_train, y_test = train_test_split(
            dataset.data, dataset.target, random_state=42
        )

    """
    module_path = dirname(__file__)
    raw_data = pd.read_csv(join(module_path, 'data', 'bankmarketing.csv.gz'), delimiter=';', compression='gzip')
    with open(join(module_path, 'descriptions', 'bankmarketing.rst'), encoding='utf-8') as f:
        description = f.read()

    # only use features pre-contact:
    # 1 - age (numeric)
    # 2 - job : type of job (categorical: "admin.","unknown","unemployed","management","housemaid","entrepreneur",
    #                        "student","blue-collar","self-employed","retired","technician","services")
    # 3 - marital : marital status (categorical: "married","divorced","single";
    #                               note: "divorced" means divorced or widowed)
    # 4 - education (categorical: "unknown","secondary","primary","tertiary")
    # 5 - default: has credit in default? (binary: "yes","no")
    # 6 - balance: average yearly balance, in euros (numeric)
    # 7 - housing: has housing loan? (binary: "yes","no")
    # 8 - loan: has personal loan? (binary: "yes","no")
    # 15 - previous: number of contacts performed before this campaign and for this client (numeric)
    # 16 - poutcome: outcome of the previous marketing campaign (categorical: "unknown","other","failure","success")

    # Following features exclude because are collected after the contact event
    # # related with the last contact of the current campaign:
    # 9 - contact: contact communication type (categorical: "unknown","telephone","cellular")
    # 10 - day: last contact day of the month (numeric)
    # 11 - month: last contact month of year (categorical: "jan", "feb", "mar", ..., "nov", "dec")
    # 12 - duration: last contact duration, in seconds (numeric)
    # # other attributes:
    # 13 - campaign: number of contacts performed during this campaign and for this client
    # 14 - pdays: number of days that passed by after the client was last contacted from a
    #       previous campaign (numeric, -1 means client was not previously contacted)

    # Filter if balance>0
    raw_data = raw_data.loc[raw_data['balance'] > 0]
    target = (raw_data.y.to_numpy() == 'yes').astype(np.int8)
    data = raw_data[
        ['age', 'balance', 'previous', 'job', 'marital', 'education', 'default', 'housing', 'loan', 'poutcome']
    ]

    fp_cost = contact_cost
    fn_cost = np.maximum(data['balance'].to_numpy() * interest_rate * term_deposit_fraction, contact_cost)
    tp_cost = contact_cost
    tn_cost = 0.0

    data.loc[:, 'default'] = data['default'].map({'yes': 1, 'no': 0})
    data.loc[:, 'housing'] = data['housing'].map({'yes': 1, 'no': 0})
    data.loc[:, 'loan'] = data['loan'].map({'yes': 1, 'no': 0})

    data = data.astype({
        'age': np.uint8,
        'balance': np.int32,
        'previous': np.uint8,
        'job': 'category',
        'marital': 'category',
        'education': 'category',
        'default': np.uint8,
        'housing': np.uint8,
        'loan': np.uint8,
        'poutcome': 'category',
    })

    data = data.rename(
        columns={
            'poutcome': 'previous_outcome',
            'loan': 'has_personal_loan',
            'housing': 'has_housing_loan',
            'default': 'has_credit_in_default',
        }
    )

    if return_X_y_costs:
        if as_frame:
            return (
                data,
                pd.Series(target, name='subscription'),
                tp_cost,
                fp_cost,
                tn_cost,
                pd.Series(fn_cost, name='fn_cost'),
            )
        else:
            return (data.to_numpy(), target, tp_cost, fp_cost, tn_cost, fn_cost)
    else:
        target_names = ['no subscription', 'subscription']
        return Dataset(
            data=data.to_numpy() if not as_frame else data,
            target=target if not as_frame else pd.Series(target, name='subscription'),
            tp_cost=tp_cost,
            fp_cost=fp_cost,
            tn_cost=tn_cost,
            fn_cost=fn_cost if not as_frame else pd.Series(fn_cost, name='fn_cost'),
            feature_names=data.columns.to_numpy() if not as_frame else data.columns,
            target_names=np.array(target_names) if not as_frame else pd.Series(target_names, name='target'),
            name='Bank Telemarketing',
            DESCR=description,
        )


@overload
def load_give_me_some_credit(
    *,
    as_frame: Literal[True],
    return_X_y_costs: Literal[False] = False,
    interest_rate: float = 0.0479,
    fund_cost: float = 0.0294,
    max_credit_line: float = 25000,
    loss_given_default: float = 0.75,
    term_length_months: int = 24,
    loan_to_income_ratio: float = 3,
) -> Dataset[pd.DataFrame, pd.Series]: ...


@overload
def load_give_me_some_credit(
    *,
    as_frame: Literal[False] = False,
    return_X_y_costs: Literal[False] = False,
    interest_rate: float = 0.0479,
    fund_cost: float = 0.0294,
    max_credit_line: float = 25000,
    loss_given_default: float = 0.75,
    term_length_months: int = 24,
    loan_to_income_ratio: float = 3,
) -> Dataset[FloatNDArray, FloatNDArray]: ...


@overload
def load_give_me_some_credit(
    *,
    as_frame: Literal[True],
    return_X_y_costs: Literal[True],
    interest_rate: float = 0.0479,
    fund_cost: float = 0.0294,
    max_credit_line: float = 25000,
    loss_given_default: float = 0.75,
    term_length_months: int = 24,
    loan_to_income_ratio: float = 3,
) -> tuple[pd.DataFrame, pd.Series, pd.Series, pd.Series, pd.Series, pd.Series]: ...


@overload
def load_give_me_some_credit(
    *,
    as_frame: Literal[False] = False,
    return_X_y_costs: Literal[True],
    interest_rate: float = 0.0479,
    fund_cost: float = 0.0294,
    max_credit_line: float = 25000,
    loss_given_default: float = 0.75,
    term_length_months: int = 24,
    loan_to_income_ratio: float = 3,
) -> tuple[FloatNDArray, FloatNDArray, FloatNDArray, FloatNDArray, FloatNDArray, FloatNDArray]: ...


def load_give_me_some_credit(
    *,
    as_frame: bool = False,
    return_X_y_costs: bool = False,
    interest_rate: float = 0.0479,
    fund_cost: float = 0.0294,
    max_credit_line: float = 25000,
    loss_given_default: float = 0.75,
    term_length_months: int = 24,
    loan_to_income_ratio: float = 3,
) -> (
    Dataset[pd.DataFrame, pd.Series]
    | Dataset[FloatNDArray, FloatNDArray]
    | tuple[pd.DataFrame, pd.Series, pd.Series, pd.Series, pd.Series, pd.Series]
    | tuple[FloatNDArray, FloatNDArray, FloatNDArray, FloatNDArray, FloatNDArray, FloatNDArray]
):
    """
    Load the "Give Me Some Credit" Kaggle credit scoring competition dataset (binary classification).

    The goal is to predict whether a customer will default on a loan in the next two years.
    The target variable is whether the customer defaulted, 'yes' = 1 and 'no' = 0.

    Only customers with a positive monthly income and a debt ratio less than 1 are considered.

    For a full data description and additional information about the dataset,
    consult the :ref:`User Guide <give_me_some_credit>`.

    =================   ==============
    Classes                          2
    Defaulters                    7616
    Non-defaulters              105299
    Samples                     112915
    Features                        10
    =================   ==============

    Parameters
    ----------
    as_frame : bool, default=False
        If True, the output will be a pandas DataFrames or Series instead of numpy arrays.
    return_X_y_costs : bool, default=False
        If True, return (data, target, tp_cost, fp_cost, tn_cost, fn_cost) instead of a Dataset object.
    interest_rate : float, default=0.02463333
        Annual interest rate of the term deposit.
    fund_cost : float, default=0.0294
        Annual cost of funds.
    max_credit_line : float, default=25000
        The maximum amount a client can borrow.
    loss_given_default : float, default=0.75
        The fraction of the loan amount which is lost if the client defaults.
    term_length_months : int, default=24
        The length of the loan term in months.
    loan_to_income_ratio : float, default=3
        The ratio of the loan amount to the client's income.

    Returns
    -------
    dataset : :class:`~empulse.datasets.Dataset` or tuple of (data, target, tp_cost, fp_cost, tn_cost, fn_cost)
        Returns a Dataset object if `return_X_y_costs=False` (default), otherwise a tuple.

    Notes
    -----
    Cost matrix

    .. list-table::

        * -
          - Actual positive :math:`y_i = 1`
          - Actual negative :math:`y_i = 0`
        * - Predicted positive :math:`\\hat{y}_i = 1`
          - ``tp_cost`` :math:`= 0`
          - ``fp_cost`` :math:`= r_i + -\\bar{r} \\cdot \\pi_0 + \\bar{Cl} \\cdot L_{gd} \\cdot \\pi_1`
        * - Predicted negative :math:`\\hat{y}_i = 0`
          - ``fn_cost`` :math:`= Cl_i \\cdot L_{gd}`
          - ``tn_cost`` :math:`= 0`

    with
        - :math:`r_i` : loss in profit by rejecting what would have been a good loan
        - :math:`\\bar{r}` : average loss in profit by rejecting what would have been a good loan
        - :math:`\\pi_0` : percentage of defaulters
        - :math:`\\pi_1` : percentage of non-defaulters
        - :math:`Cl_i` : credit line of the client
        - :math:`\\bar{Cl}` : average credit line
        - :math:`L_{gd}` : the fraction of the loan amount which is lost if the client defaults

    References
    ----------
    .. [1] A. Correa Bahnsen, D.Aouada, B, Ottersten,
           "Example-Dependent Cost-Sensitive Logistic Regression for Credit Scoring",
           in Proceedings of the International Conference on Machine Learning and Applications, 2014.

    Examples
    --------

    .. code-block:: python

        from empulse.datasets import load_give_me_some_credit
        from sklearn.model_selection import train_test_split

        dataset = load_give_me_some_credit()
        X_train, X_test, y_train, y_test = train_test_split(
            dataset.data, dataset.target, random_state=42
        )

    """
    module_path = dirname(__file__)
    raw_data = pd.read_csv(join(module_path, 'data', 'creditscoring1.csv.gz'), delimiter=',', compression='gzip')
    with open(join(module_path, 'descriptions', 'creditscoring1.rst'), encoding='utf-8') as f:
        description = f.read()

    # Exclude MonthlyIncome = nan or =0 or DebtRatio >1
    raw_data = raw_data.dropna()
    raw_data = raw_data.loc[(raw_data['MonthlyIncome'] > 0)]
    raw_data = raw_data.loc[(raw_data['DebtRatio'] < 1)]

    target = raw_data['SeriousDlqin2yrs'].to_numpy().astype(np.int64)

    data = raw_data.drop(['SeriousDlqin2yrs', 'id'], axis=1)

    # Calculate cost_mat (see[1])
    cost_mat_parameters = {
        'int_r': interest_rate / 12,
        'int_cf': fund_cost / 12,
        'cl_max': max_credit_line,
        'n_term': term_length_months,
        'k': loan_to_income_ratio,
        'lgd': loss_given_default,
    }

    pi_1 = target.mean()
    # cost_mat[FP,FN,TP,TN]
    cost_mat = _creditscoring_costmat(
        data['MonthlyIncome'].to_numpy(), data['DebtRatio'].to_numpy(), pi_1, cost_mat_parameters
    )

    # unroll into separate costs
    fp_cost = cost_mat[:, 0]
    fn_cost = cost_mat[:, 1]

    # normalize feature names
    column_mapping = {
        'RevolvingUtilizationOfUnsecuredLines': 'revolving_utilization',
        'age': 'age',
        'NumberOfTime30-59DaysPastDueNotWorse': 'n_times_late_30_59_days',
        'DebtRatio': 'debt_ratio',
        'MonthlyIncome': 'monthly_income',
        'NumberOfOpenCreditLinesAndLoans': 'n_open_credit_lines',
        'NumberOfTimes90DaysLate': 'n_times_late_over_90_days',
        'NumberRealEstateLoansOrLines': 'n_real_estate_loans',
        'NumberOfTime60-89DaysPastDueNotWorse': 'n_times_late_60_89_days',
        'NumberOfDependents': 'n_dependents',
    }

    data = data.rename(columns=column_mapping)

    new_order = [
        'monthly_income',
        'debt_ratio',
        'revolving_utilization',
        'age',
        'n_dependents',
        'n_open_credit_lines',
        'n_real_estate_loans',
        'n_times_late_30_59_days',
        'n_times_late_60_89_days',
        'n_times_late_over_90_days',
    ]

    # Reorder columns
    data = data.reindex(columns=new_order)

    data = data.astype({
        'monthly_income': np.float64,
        'debt_ratio': np.float64,
        'revolving_utilization': np.float64,
        'age': np.uint8,
        'n_dependents': np.uint8,
        'n_open_credit_lines': np.uint8,
        'n_real_estate_loans': np.uint8,
        'n_times_late_30_59_days': np.uint8,
        'n_times_late_60_89_days': np.uint8,
        'n_times_late_over_90_days': np.uint8,
    })

    if return_X_y_costs:
        if as_frame:
            return (
                data,
                pd.Series(target, name='default'),
                0.0,
                pd.Series(fp_cost, name='fp_cost'),
                0.0,
                pd.Series(fn_cost, name='fn_cost'),
            )
        else:
            return (data.to_numpy(), target, 0.0, fp_cost, 0.0, fn_cost)
    else:
        target_names = ['no default', 'default']
        return Dataset(
            data=data.to_numpy() if not as_frame else data,
            target=target if not as_frame else pd.Series(target, name='default'),
            tp_cost=0.0,
            fp_cost=fp_cost if not as_frame else pd.Series(fp_cost, name='fp_cost'),
            tn_cost=0.0,
            fn_cost=fn_cost if not as_frame else pd.Series(fn_cost, name='fn_cost'),
            feature_names=data.columns.to_numpy() if not as_frame else data.columns,
            target_names=np.array(target_names) if not as_frame else pd.Series(target_names, name='target'),
            name='Give Me Some Credit',
            DESCR=description,
        )


@overload
def load_credit_scoring_pakdd(
    *,
    as_frame: Literal[True],
    return_X_y_costs: Literal[False] = False,
    interest_rate: float = 0.0479,
    fund_cost: float = 0.0294,
    max_credit_line: float = 25000,
    loss_given_default: float = 0.75,
    term_length_months: int = 24,
    loan_to_income_ratio: float = 3,
) -> Dataset[pd.DataFrame, pd.Series]: ...


@overload
def load_credit_scoring_pakdd(
    *,
    as_frame: Literal[False] = False,
    return_X_y_costs: Literal[False] = False,
    interest_rate: float = 0.0479,
    fund_cost: float = 0.0294,
    max_credit_line: float = 25000,
    loss_given_default: float = 0.75,
    term_length_months: int = 24,
    loan_to_income_ratio: float = 3,
) -> Dataset[FloatNDArray, FloatNDArray]: ...


@overload
def load_credit_scoring_pakdd(
    *,
    as_frame: Literal[True],
    return_X_y_costs: Literal[True],
    interest_rate: float = 0.0479,
    fund_cost: float = 0.0294,
    max_credit_line: float = 25000,
    loss_given_default: float = 0.75,
    term_length_months: int = 24,
    loan_to_income_ratio: float = 3,
) -> tuple[pd.DataFrame, pd.Series, pd.Series, pd.Series, pd.Series, pd.Series]: ...


@overload
def load_credit_scoring_pakdd(
    *,
    as_frame: Literal[False] = False,
    return_X_y_costs: Literal[True],
    interest_rate: float = 0.0479,
    fund_cost: float = 0.0294,
    max_credit_line: float = 25000,
    loss_given_default: float = 0.75,
    term_length_months: int = 24,
    loan_to_income_ratio: float = 3,
) -> tuple[FloatNDArray, FloatNDArray, FloatNDArray, FloatNDArray, FloatNDArray, FloatNDArray]: ...


def load_credit_scoring_pakdd(
    *,
    as_frame: bool = False,
    return_X_y_costs: bool = False,
    interest_rate: float = 0.63,
    fund_cost: float = 0.165,
    max_credit_line: float = 25000,
    loss_given_default: float = 0.75,
    term_length_months: int = 24,
    loan_to_income_ratio: float = 3,
) -> (
    Dataset[pd.DataFrame, pd.Series]
    | Dataset[FloatNDArray, FloatNDArray]
    | tuple[pd.DataFrame, pd.Series, pd.Series, pd.Series, pd.Series, pd.Series]
    | tuple[FloatNDArray, FloatNDArray, FloatNDArray, FloatNDArray, FloatNDArray, FloatNDArray]
):
    """
    Load the credit scoring PAKDD 2009 competition dataset (binary classification).

    The goal is to predict whether a customer will default on a loan in the next two years.
    The target variable is whether the customer defaulted, 'yes' = 1 and 'no' = 0.

    Only clients with a personal income between 100 and 10000 are considered.

    For a full data description and additional information about the dataset,
    consult the :ref:`User Guide <credit_scoring_pakdd>`.

    =================   ==============
    Classes                          2
    Defaulters                    7743
    Non-defaulters               31195
    Samples                      38938
    Features                        25
    =================   ==============

    Parameters
    ----------
    as_frame : bool, default=False
        If True, the output will be a pandas DataFrames or Series instead of numpy arrays.
    return_X_y_costs : bool, default=False
        If True, return (data, target, tp_cost, fp_cost, tn_cost, fn_cost) instead of a Dataset object.
    interest_rate : float, default=0.63
        Annual interest rate of the term deposit.
    fund_cost : float, default=0.165
        Annual cost of funds.
    max_credit_line : float, default=25000
        The maximum amount a client can borrow.
    loss_given_default : float, default=0.75
        The amount of the loan amount which is lost if the client defaults.
    term_length_months : int, default=24
        The length of the loan term in months.
    loan_to_income_ratio : float, default=3
        The ratio of the loan amount to the client's income.

    Returns
    -------
    dataset : :class:`~empulse.datasets.Dataset` or tuple of (data, target, tp_cost, fp_cost, tn_cost, fn_cost)
        Returns a Dataset object if `return_X_y_costs=False` (default), otherwise a tuple.

    Notes
    -----
    Cost matrix

    .. list-table::

        * -
          - Actual positive :math:`y_i = 1`
          - Actual negative :math:`y_i = 0`
        * - Predicted positive :math:`\\hat{y}_i = 1`
          - ``tp_cost`` :math:`= 0`
          - ``fp_cost`` :math:`= r_i + -\\bar{r} \\cdot \\pi_0 + \\bar{Cl} \\cdot L_{gd} \\cdot \\pi_1`
        * - Predicted negative :math:`\\hat{y}_i = 0`
          - ``fn_cost`` :math:`= Cl_i \\cdot L_{gd}`
          - ``tn_cost`` :math:`= 0`

    with
        - :math:`r_i` : loss in profit by rejecting what would have been a good loan
        - :math:`\\bar{r}` : average loss in profit by rejecting what would have been a good loan
        - :math:`\\pi_0` : percentage of defaulters
        - :math:`\\pi_1` : percentage of non-defaulters
        - :math:`Cl_i` : credit line of the client
        - :math:`\\bar{Cl}` : average credit line
        - :math:`L_{gd}` : the fraction of the loan amount which is lost if the client defaults

    References
    ----------
    .. [1] A. Correa Bahnsen, D.Aouada, B, Ottersten,
           "Example-Dependent Cost-Sensitive Logistic Regression for Credit Scoring",
           in Proceedings of the International Conference on Machine Learning and Applications,
           , 2014.

    Examples
    --------

    .. code-block:: python

        from empulse.datasets import load_credit_scoring_pakdd
        from sklearn.model_selection import train_test_split

        dataset = load_credit_scoring_pakdd()
        X_train, X_test, y_train, y_test = train_test_split(
            dataset.data, dataset.target, random_state=42
        )

    """
    module_path = dirname(__file__)
    raw_data = pd.read_csv(join(module_path, 'data', 'creditscoring2.csv.gz'), delimiter='\t', compression='gzip')
    with open(join(module_path, 'descriptions', 'creditscoring2.rst'), encoding='utf-8') as f:
        description = f.read()

    # Exclude TARGET_LABEL_BAD=1 == 'N'
    raw_data = raw_data.loc[raw_data['TARGET_LABEL_BAD=1'] != 'N']

    # Exclude 100<PERSONAL_NET_INCOME<10000
    raw_data = raw_data.loc[(raw_data['PERSONAL_NET_INCOME'].to_numpy().astype(np.float64) > 100)]
    raw_data = raw_data.loc[(raw_data['PERSONAL_NET_INCOME'].to_numpy().astype(np.float64) < 10000)]

    target = raw_data['TARGET_LABEL_BAD=1'].to_numpy().astype(np.int64)
    data = raw_data.drop(['TARGET_LABEL_BAD=1'], axis=1)

    # drop the last column
    data = data.iloc[:, :-1]

    continuous_columns = ['MATE_INCOME', 'PERSONAL_NET_INCOME']
    integer_columns = [
        'ID_SHOP',
        'AGE',
        'AREA_CODE_RESIDENCIAL_PHONE',
        'PAYMENT_DAY',
        'SHOP_RANK',
        'MONTHS_IN_RESIDENCE',
        'MONTHS_IN_THE_JOB',
        'PROFESSION_CODE',
        'QUANT_ADDITIONAL_CARDS_IN_THE_APPLICATION',
    ]
    data[continuous_columns] = data[continuous_columns].astype('float64')
    data[integer_columns] = data[integer_columns].astype('int32')

    # Calculate cost_mat (see[1])
    cost_mat_parameters = {
        'int_r': interest_rate / 12,
        'int_cf': fund_cost / 12,
        'cl_max': max_credit_line * 0.33,
        'n_term': term_length_months,
        'k': loan_to_income_ratio,
        'lgd': loss_given_default,
    }

    n_samples = data.shape[0]
    pi_1 = target.mean()
    monthly_income = data['PERSONAL_NET_INCOME'].to_numpy() * 0.33
    cost_mat = _creditscoring_costmat(monthly_income, np.zeros(n_samples), pi_1, cost_mat_parameters)

    # unroll into separate costs
    fp_cost = cost_mat[:, 0]
    fn_cost = cost_mat[:, 1]

    # convert all columns which start with flag to integers
    flag_columns = [col for col in data.columns if col.startswith('FLAG')]
    data[flag_columns] = (data[flag_columns] == 'Y').astype(np.int8)

    # normalize feature names
    data.columns = data.columns.str.lower().str.replace('#', '').str.replace('quant', 'n')
    data.columns = data.columns.str.replace('_in_the_application', '').str.replace('residencial', 'residential')
    data.loc[:, 'sex'] = data['sex'].map({'M': 1, 'F': 0})
    # fill in single  missing value as male
    data.loc[:, 'sex'] = data.sex.fillna('1')

    column_mapping = {
        'flag_residence_town_eq_working_town': 'lives_in_work_town',
        'flag_residence_state_eq_working_state': 'lives_in_work_state',
        'flag_residential_address_eq_postal_address': 'has_same_postal_address',
        'flag_residential_phone': 'has_residential_phone',
        'sex': 'is_male',
        'flag_mothers_name': 'filled_in_mothers_name',
        'flag_fathers_name': 'filled_in_fathers_name',
        'mate_income': 'partner_income',
        'flag_other_card': 'has_other_card',
        'flag_mobile_phone': 'has_mobile_phone',
        'flag_contact_phone': 'has_contact_phone',
        'cod_application_booth': 'application_booth_code',
        'flag_card_insurance_option': 'has_card_insurance',
        'id_shop': 'shop_code',
    }
    data = data.rename(columns=column_mapping)

    # remap values of matiral_status to more readable values
    data.loc[:, 'marital_status'] = data['marital_status'].map({
        'S': 'single',
        'M': 'married',
        'D': 'divorced',
        'W': 'widowed',
        'O': 'other',
    })
    # remap values of residence_type to more readable values
    data.loc[:, 'residence_type'] = data['residence_type'].map({
        'P': 'owned',
        'A': 'rented',
        'C': 'parents',
        'O': 'other',
    })

    # Desired column order
    new_order = [
        'age',
        'personal_net_income',
        'partner_income',
        'months_in_residence',
        'months_in_the_job',
        'payment_day',
        'n_banking_accounts',
        'n_additional_cards',
        'is_male',
        'has_residential_phone',
        'has_mobile_phone',
        'has_contact_phone',
        'has_same_postal_address',
        'has_other_card',
        'has_card_insurance',
        'lives_in_work_town',
        'lives_in_work_state',
        'filled_in_mothers_name',
        'filled_in_fathers_name',
        'shop_rank',
        'marital_status',
        'residence_type',
        'area_code_residential_phone',
        'shop_code',
        'application_booth_code',
        'profession_code',
    ]

    # Reorder columns
    data = data.reindex(columns=new_order)

    data = data.astype({
        'age': np.uint8,
        'personal_net_income': np.float32,
        'partner_income': np.float32,
        'months_in_residence': np.uint8,
        'months_in_the_job': np.uint8,
        'payment_day': np.uint8,
        'n_banking_accounts': np.uint8,
        'n_additional_cards': np.uint8,
        'is_male': np.uint8,
        'has_residential_phone': np.uint8,
        'has_mobile_phone': np.uint8,
        'has_contact_phone': np.uint8,
        'has_same_postal_address': np.uint8,
        'has_other_card': np.uint8,
        'has_card_insurance': np.uint8,
        'lives_in_work_town': np.uint8,
        'lives_in_work_state': np.uint8,
        'filled_in_mothers_name': np.uint8,
        'filled_in_fathers_name': np.uint8,
        'shop_rank': 'category',
        'marital_status': 'category',
        'residence_type': 'category',
        'area_code_residential_phone': 'category',
        'shop_code': 'category',
        'application_booth_code': 'category',
        'profession_code': 'category',
    })

    # remove flag_card_insurance_option
    data = data.drop(['has_card_insurance'], axis=1)

    if return_X_y_costs:
        if as_frame:
            return (
                data,
                pd.Series(target, name='default'),
                0.0,
                pd.Series(fp_cost, name='fp_cost'),
                0.0,
                pd.Series(fn_cost, name='fn_cost'),
            )
        else:
            return (data.to_numpy(), target, 0.0, fp_cost, 0.0, fn_cost)
    else:
        target_names = ['no default', 'default']
        return Dataset(
            data=data.to_numpy() if not as_frame else data,
            target=target if not as_frame else pd.Series(target, name='default'),
            tp_cost=0.0,
            fp_cost=fp_cost if not as_frame else pd.Series(fp_cost, name='fp_cost'),
            tn_cost=0.0,
            fn_cost=fn_cost if not as_frame else pd.Series(fn_cost, name='fn_cost'),
            feature_names=data.columns.to_numpy() if not as_frame else data.columns,
            target_names=np.array(target_names) if not as_frame else pd.Series(target_names, name='target'),
            name='Credit Scoring PAKDD 2009',
            DESCR=description,
        )


def _creditscoring_costmat(
    income: FloatNDArray, debt: FloatNDArray, pi_1: float, cost_mat_parameters: dict[str, float]
) -> FloatNDArray:
    """Private function to calculate the cost matrix of credit scoring models.

    Parameters
    ----------
    income : array of shape = [n_samples]
        Monthly income of each example

    debt : array of shape = [n_samples]
        Debt ratio each example

    pi_1 : float
        Percentage of positives in the training set

    References
    ----------
    .. [1] A. Correa Bahnsen, D.Aouada, B, Ottersten,
           "Example-Dependent Cost-Sensitive Logistic Regression for Credit Scoring",
           in Proceedings of the International Conference on Machine Learning and Applications,
           , 2014.

    Returns
    -------
    cost_mat : array-like of shape = [n_samples, 4]
        Cost matrix of the classification problem
        Where the columns represent the costs of: false positives, false negatives,
        true positives and true negatives, for each example.
    """

    def calculate_a(cl_i: float, int_: float, n_term: float) -> float:
        return float(cl_i * ((int_ * (1 + int_) ** n_term) / ((1 + int_) ** n_term - 1)))

    def calculate_pv(a: float, int_: float, n_term: float) -> float:
        return float(a / int_ * (1 - 1 / (1 + int_) ** n_term))

    # Calculate credit line Cl
    def calculate_cl(k: float, inc_i: float, cl_max: float, debt_i: float, int_r: float, n_term: float) -> float:
        cl_k = k * inc_i
        a = calculate_a(cl_k, int_r, n_term)
        cl_debt = calculate_pv(inc_i * min(a / inc_i, 1 - debt_i), int_r, n_term)
        return min(cl_k, cl_max, cl_debt)

    # calculate costs
    def calculate_cost_fn(cl_i: float, lgd: float) -> float:
        return cl_i * lgd

    def calculate_cost_fp(
        cl_i: float, int_r: float, n_term: float, int_cf: float, pi_1: float, lgd: float, cl_avg: float
    ) -> float:
        a = calculate_a(cl_i, int_r, n_term)
        pv = calculate_pv(a, int_cf, n_term)
        r = pv - cl_i
        r_avg = calculate_pv(calculate_a(cl_avg, int_r, n_term), int_cf, n_term) - cl_avg
        cost_fp = r - (1 - pi_1) * r_avg + pi_1 * calculate_cost_fn(cl_avg, lgd)
        return max(0, cost_fp)

    v_calculate_cost_fp = np.vectorize(calculate_cost_fp)
    v_calculate_cost_fn = np.vectorize(calculate_cost_fn)

    v_calculate_cl = np.vectorize(calculate_cl)

    # Parameters
    k = cost_mat_parameters['k']
    int_r = cost_mat_parameters['int_r']
    n_term = cost_mat_parameters['n_term']
    int_cf = cost_mat_parameters['int_cf']
    lgd = cost_mat_parameters['lgd']
    cl_max = cost_mat_parameters['cl_max']

    cl = v_calculate_cl(k, income, cl_max, debt, int_r, n_term)
    cl_avg = cl.mean()

    n_samples = income.shape[0]
    cost_mat = np.zeros((n_samples, 4))  # cost_mat[FP,FN,TP,TN]
    cost_mat[:, 0] = v_calculate_cost_fp(cl, int_r, n_term, int_cf, pi_1, lgd, cl_avg)
    cost_mat[:, 1] = v_calculate_cost_fn(cl, lgd)
    cost_mat[:, 2] = 0.0
    cost_mat[:, 3] = 0.0

    return cost_mat



import sys
import threading
from collections.abc import Callable
from numbers import Integral, Real
from typing import Any, ClassVar, Literal

import numpy as np
from joblib import Parallel, delayed
from scipy.sparse import csr_matrix, issparse
from sklearn import config_context
from sklearn.base import BaseEstimator, ClassifierMixin, _fit_context, clone
from sklearn.ensemble import BaggingClassifier, RandomForestClassifier
from sklearn.ensemble._base import _partition_estimators
from sklearn.tree import DecisionTreeClassifier
from sklearn.utils._available_if import available_if
from sklearn.utils._mask import indices_to_mask
from sklearn.utils._param_validation import StrOptions
from sklearn.utils.validation import _estimator_has, check_is_fitted, check_random_state

from ..._common import Parameter
from ..._types import FloatArrayLike, FloatNDArray, IntArrayLike, IntNDArray, ParameterConstraint
from ...metrics import Metric, expected_cost_loss
from ...utils._sklearn_compat import Tags, type_of_target, validate_data  # type: ignore[attr-defined]
from ._cs_mixin import CostSensitiveMixin
from ._impurity import CostImpurity, EntropyCostImpurity, GiniCostImpurity
from .cstree import CSTreeClassifier

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

RF_PARAM_CONSTRAINTS = RandomForestClassifier._parameter_constraints.copy()
RF_PARAM_CONSTRAINTS.pop('criterion')


class CSForestClassifier(CostSensitiveMixin, ClassifierMixin, BaseEstimator):
    """
    Cost-sensitive random forest classifier.

    Parameters
    ----------
    n_estimators : int, default=100
        The number of trees in the forest.

    tp_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of true positives. If ``float``, then all true positives have the same cost.
        If array-like, then it is the cost of each true positive classification.
        Is overwritten if another `tp_cost` is passed to the ``fit`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit`` method.

    fp_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of false positives. If ``float``, then all false positives have the same cost.
        If array-like, then it is the cost of each false positive classification.
        Is overwritten if another `fp_cost` is passed to the ``fit`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit`` method.

    tn_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of true negatives. If ``float``, then all true negatives have the same cost.
        If array-like, then it is the cost of each true negative classification.
        Is overwritten if another `tn_cost` is passed to the ``fit`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit`` method.

    fn_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of false negatives. If ``float``, then all false negatives have the same cost.
        If array-like, then it is the cost of each false negative classification.
        Is overwritten if another `fn_cost` is passed to the ``fit`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit`` method.

    n_estimators : int, default=100
        The number of trees in the forest.

        .. versionchanged:: 0.22
           The default value of ``n_estimators`` changed from 10 to 100
           in 0.22.

    loss : Metric or None, default=None
        The metric to measure the quality of a split.
        If None, the cost impurity is used.

    criterion : {"cost",, "gini", "log_loss" or "entropy"}, default="cost"
        The function to measure the quality of a split.

        How the measure to estimate quality of a split is weighted.

        - If ``"cost"``: The metric is used normally, without extra weighting.
        - If ``"gini"``: The Gini impurity is used to weight the metric.
        - If ``"log_loss"`` or ``"entropy"``: The Shannon information gain is used to weight the metric.

    combination : {"majority_voting', 'weighted_voting'}, default="majority_voting"
        How to combine the predictions of the individual models.

        - "majority_voting": the majority vote of the models.
        - "weighted_voting": the models are weighted by their oob score calculates with the ....

    max_depth : int, default=None
        The maximum depth of the tree. If None, then nodes are expanded until
        all leaves are pure or until all leaves contain less than
        min_samples_split samples.

    min_samples_split : int or float, default=2
        The minimum number of samples required to split an internal node:

        - If int, then consider `min_samples_split` as the minimum number.
        - If float, then `min_samples_split` is a fraction and
          `ceil(min_samples_split * n_samples)` are the minimum
          number of samples for each split.

        .. versionchanged:: 0.18
           Added float values for fractions.

    min_samples_leaf : int or float, default=1
        The minimum number of samples required to be at a leaf node.
        A split point at any depth will only be considered if it leaves at
        least ``min_samples_leaf`` training samples in each of the left and
        right branches.  This may have the effect of smoothing the model,
        especially in regression.

        - If int, then consider `min_samples_leaf` as the minimum number.
        - If float, then `min_samples_leaf` is a fraction and
          `ceil(min_samples_leaf * n_samples)` are the minimum
          number of samples for each node.

    min_weight_fraction_leaf : float, default=0.0
        The minimum weighted fraction of the sum total of weights (of all
        the input samples) required to be at a leaf node. Samples have
        equal weight when sample_weight is not provided.

    max_features : {"sqrt", "log2", None}, int or float, default="sqrt"
        The number of features to consider when looking for the best split:

        - If int, then consider `max_features` features at each split.
        - If float, then `max_features` is a fraction and
          `max(1, int(max_features * n_features_in_))` features are considered at each
          split.
        - If "sqrt", then `max_features=sqrt(n_features)`.
        - If "log2", then `max_features=log2(n_features)`.
        - If None, then `max_features=n_features`.

        Note: the search for a split does not stop until at least one
        valid partition of the node samples is found, even if it requires to
        effectively inspect more than ``max_features`` features.

    max_leaf_nodes : int, default=None
        Grow trees with ``max_leaf_nodes`` in best-first fashion.
        Best nodes are defined as relative reduction in impurity.
        If None then unlimited number of leaf nodes.

    min_impurity_decrease : float, default=0.0
        A node will be split if this split induces a decrease of the impurity
        greater than or equal to this value.

        The weighted impurity decrease equation is the following::

            N_t / N * (impurity - N_t_R / N_t * right_impurity
                                - N_t_L / N_t * left_impurity)

        where ``N`` is the total number of samples, ``N_t`` is the number of
        samples at the current node, ``N_t_L`` is the number of samples in the
        left child, and ``N_t_R`` is the number of samples in the right child.

        ``N``, ``N_t``, ``N_t_R`` and ``N_t_L`` all refer to the weighted sum,
        if ``sample_weight`` is passed.

    bootstrap : bool, default=True
        Whether bootstrap samples are used when building trees. If False, the
        whole dataset is used to build each tree.

    oob_score : bool or callable, default=False
        Whether to use out-of-bag samples to estimate the generalization score.
        By default, :func:`~sklearn.metrics.accuracy_score` is used.
        Provide a callable with signature `metric(y_true, y_pred)` to use a
        custom metric. Only available if `bootstrap=True`.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`fit`, :meth:`predict`,
        :meth:`decision_path` and :meth:`apply` are all parallelized over the
        trees. ``None`` means 1 unless in a :obj:`joblib.parallel_backend`
        context. ``-1`` means using all processors. See :term:`Glossary
        <n_jobs>` for more details.

    random_state : int, RandomState instance or None, default=None
        Controls both the randomness of the bootstrapping of the samples used
        when building trees (if ``bootstrap=True``) and the sampling of the
        features to consider when looking for the best split at each node
        (if ``max_features < n_features``).
        See :term:`Glossary <random_state>` for details.

    verbose : int, default=0
        Controls the verbosity when fitting and predicting.

    warm_start : bool, default=False
        When set to ``True``, reuse the solution of the previous call to fit
        and add more estimators to the ensemble, otherwise, just fit a whole
        new forest. See :term:`Glossary <warm_start>` and
        :ref:`tree_ensemble_warm_start` for details.

    class_weight : {"balanced", "balanced_subsample"}, dict or list of dicts, \
            default=None
        Weights associated with classes in the form ``{class_label: weight}``.
        If not given, all classes are supposed to have weight one. For
        multi-output problems, a list of dicts can be provided in the same
        order as the columns of y.

        Note that for multioutput (including multilabel) weights should be
        defined for each class of every column in its own dict. For example,
        for four-class multilabel classification weights should be
        [{0: 1, 1: 1}, {0: 1, 1: 5}, {0: 1, 1: 1}, {0: 1, 1: 1}] instead of
        [{1:1}, {2:5}, {3:1}, {4:1}].

        The "balanced" mode uses the values of y to automatically adjust
        weights inversely proportional to class frequencies in the input data
        as ``n_samples / (n_classes * np.bincount(y))``

        The "balanced_subsample" mode is the same as "balanced" except that
        weights are computed based on the bootstrap sample for every tree
        grown.

        For multi-output, the weights of each column of y will be multiplied.

        Note that these weights will be multiplied with sample_weight (passed
        through the fit method) if sample_weight is specified.

    ccp_alpha : non-negative float, default=0.0
        Complexity parameter used for Minimal Cost-Complexity Pruning. The
        subtree with the largest cost complexity that is smaller than
        ``ccp_alpha`` will be chosen. By default, no pruning is performed. See
        :ref:`minimal_cost_complexity_pruning` for details. See
        :ref:`sphx_glr_auto_examples_tree_plot_cost_complexity_pruning.py`
        for an example of such pruning.

    max_samples : int or float, default=None
        If bootstrap is True, the number of samples to draw from X
        to train each base estimator.

        - If None (default), then draw `X.shape[0]` samples.
        - If int, then draw `max_samples` samples.
        - If float, then draw `max(round(n_samples * max_samples), 1)` samples. Thus,
          `max_samples` should be in the interval `(0.0, 1.0]`.

    monotonic_cst : array-like of int of shape (n_features), default=None
        Indicates the monotonicity constraint to enforce on each feature.
          - 1: monotonic increase
          - 0: no constraint
          - -1: monotonic decrease

        If monotonic_cst is None, no constraints are applied.

        Monotonicity constraints are not supported for:
          - multiclass classifications (i.e. when `n_classes > 2`),
          - multioutput classifications (i.e. when `n_outputs_ > 1`),
          - classifications trained on data with missing values.

        The constraints hold over the probability of the positive class.

        Read more in the :ref:`User Guide <monotonic_cst_gbdt>`.

    Attributes
    ----------
    estimator_ : :class:`~sklearn.tree.RandomForestClassifier`
        The underlying RandomForestClassifier estimator.

    estimators_ : list of DecisionTreeClassifier
        The collection of fitted sub-estimators.

    classes_ : ndarray of shape (n_classes,) or a list of such arrays
        The classes labels (single output problem), or a list of arrays of
        class labels (multi-output problem).

    n_classes_ : int or list
        The number of classes (single output problem), or a list containing the
        number of classes for each output (multi-output problem).

    n_features_in_ : int
        Number of features seen during :term:`fit`.

        .. versionadded:: 0.24

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during :term:`fit`. Defined only when `X`
        has feature names that are all strings.

        .. versionadded:: 1.0

    n_outputs_ : int
        The number of outputs when ``fit`` is performed.

    feature_importances_ : ndarray of shape (n_features,)
        The impurity-based feature importances.
        The higher, the more important the feature.
        The importance of a feature is computed as the (normalized)
        total reduction of the criterion brought by that feature.  It is also
        known as the Gini importance.

        Warning: impurity-based feature importances can be misleading for
        high cardinality features (many unique values). See
        :func:`sklearn.inspection.permutation_importance` as an alternative.

    oob_score_ : float
        Score of the training dataset obtained using an out-of-bag estimate.
        This attribute exists only when ``oob_score`` is True.

    oob_decision_function_ : ndarray of shape (n_samples, n_classes) or \
            (n_samples, n_classes, n_outputs)
        Decision function computed with out-of-bag estimate on the training
        set. If n_estimators is small it might be possible that a data point
        was never left out during the bootstrap. In this case,
        `oob_decision_function_` might contain NaN. This attribute exists
        only when ``oob_score`` is True.

    estimators_samples_ : list of arrays
        The subset of drawn samples (i.e., the in-bag samples) for each base
        estimator. Each subset is defined by an array of the indices selected.

    References
    ----------

    .. [1] Correa Bahnsen, A., Aouada, D., & Ottersten, B.
           `"Ensemble of Example-Dependent Cost-Sensitive Decision Trees" <http://arxiv.org/abs/1505.04637>`__,
           2015, http://arxiv.org/abs/1505.04637.
    """

    _parameter_constraints: ClassVar[ParameterConstraint] = {
        'tp_cost': ['array-like', Real],
        'tn_cost': ['array-like', Real],
        'fn_cost': ['array-like', Real],
        'fp_cost': ['array-like', Real],
        'criterion': [StrOptions({'cost', 'log_loss', 'gini', 'entropy'}), Metric],
        'loss': [Metric, None],
        'combination': [
            StrOptions({'majority_voting', 'weighted_voting'}),
        ],
        **RF_PARAM_CONSTRAINTS,
    }

    def __init__(
        self,
        n_estimators: int = 100,
        *,
        tp_cost: FloatArrayLike | float = 0.0,
        tn_cost: FloatArrayLike | float = 0.0,
        fn_cost: FloatArrayLike | float = 0.0,
        fp_cost: FloatArrayLike | float = 0.0,
        loss: Metric | None = None,
        criterion: Literal['cost', 'gini', 'entropy', 'log_loss'] = 'cost',
        combination: Literal['majority_voting', 'weighted_voting'] = 'majority_voting',
        max_depth: int | None = None,
        min_samples_split: int | float = 2,
        min_samples_leaf: int | float = 1,
        min_weight_fraction_leaf: float = 0.0,
        max_features: Literal['sqrt', 'log2'] | int | float = 'sqrt',
        max_leaf_nodes: int | None = None,
        min_impurity_decrease: float = 0.0,
        bootstrap: bool = True,
        oob_score: bool | Callable[[Any, Any], float] = False,
        n_jobs: int | None = None,
        random_state: int | np.random.RandomState | None = None,
        verbose: bool | int = 0,
        warm_start: bool = False,
        class_weight: dict[int, float] | Literal['balanced'] | None = None,
        ccp_alpha: float = 0.0,
        max_samples: int | float | None = None,
        monotonic_cst: IntArrayLike | None = None,
    ):
        self.n_estimators = n_estimators
        self.tp_cost = tp_cost
        self.tn_cost = tn_cost
        self.fn_cost = fn_cost
        self.fp_cost = fp_cost
        self.loss = loss
        self.criterion = criterion
        self.combination = combination
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes
        self.min_impurity_decrease = min_impurity_decrease
        self.bootstrap = bootstrap
        self.oob_score = oob_score
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.verbose = verbose
        self.warm_start = warm_start
        self.class_weight = class_weight
        self.ccp_alpha = ccp_alpha
        self.max_samples = max_samples
        self.monotonic_cst = monotonic_cst
        super().__init__()

    def _more_tags(self) -> dict[str, bool]:
        return {
            'binary_only': True,
            'poor_score': True,
        }

    def __sklearn_tags__(self) -> Tags:
        tags = super().__sklearn_tags__()
        tags.classifier_tags.multi_class = False
        tags.classifier_tags.poor_score = True
        return tags

    def _get_metric_loss(self) -> Metric | None:
        """Get the metric loss function if available."""
        if isinstance(self.loss, Metric):
            return self.loss
        return None

    @property
    def estimators_(self) -> list[DecisionTreeClassifier]:
        """The collection of fitted sub-estimators."""
        check_is_fitted(self)
        estimators: list[DecisionTreeClassifier] = self.estimator_.estimators_
        return estimators

    @property
    def n_classes_(self) -> int | list[int]:
        """The number of classes seen during :term:`fit`."""
        check_is_fitted(self)
        n_classes: int | list[int] = self.estimator_.n_classes_
        return n_classes

    @property
    def feature_importances_(self) -> FloatNDArray:
        """The impurity-based feature importances."""
        check_is_fitted(self)
        importances: FloatNDArray = self.estimator_.feature_importances_
        return importances

    @property
    def oob_score_(self) -> float:
        """Score of the training dataset obtained using an out-of-bag estimate."""
        check_is_fitted(self)
        oob_score: float = self.estimator_.oob_score_
        return oob_score

    @property
    def oob_decision_function_(self) -> FloatNDArray:
        """Decision function computed with out-of-bag estimate on the training set."""
        check_is_fitted(self)
        oob_decision_function: FloatNDArray = self.estimator_.oob_decision_function_
        return oob_decision_function

    @property
    def estimators_samples_(self) -> list[IntNDArray]:
        """The subset of drawn samples (i.e., the in-bag samples) for each base estimator."""
        check_is_fitted(self)
        estimators_samples: list[IntNDArray] = self.estimator_.estimators_samples_
        return estimators_samples

    @_fit_context(prefer_skip_nested_validation=True)  # type: ignore[misc]
    def fit(
        self,
        X: FloatArrayLike,
        y: IntArrayLike,
        *,
        tp_cost: FloatArrayLike | float | Parameter = Parameter.UNCHANGED,
        tn_cost: FloatArrayLike | float | Parameter = Parameter.UNCHANGED,
        fn_cost: FloatArrayLike | float | Parameter = Parameter.UNCHANGED,
        fp_cost: FloatArrayLike | float | Parameter = Parameter.UNCHANGED,
        **loss_params: Any,
    ) -> Self:
        """
        Build an example-dependent cost-sensitive decision tree from the training set.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input samples.

        y : array-like of shape (n_samples,)
            Ground truth (correct) labels.

        tp_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of true positives. If ``float``, then all true positives have the same cost.
            If array-like, then it is the cost of each true positive classification.

        fp_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of false positives. If ``float``, then all false positives have the same cost.
            If array-like, then it is the cost of each false positive classification.

        tn_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of true negatives. If ``float``, then all true negatives have the same cost.
            If array-like, then it is the cost of each true negative classification.

        fn_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of false negatives. If ``float``, then all false negatives have the same cost.
            If array-like, then it is the cost of each false negative classification.

        loss_params : dict
            Additional keyword arguments to pass to the loss function if using a custom loss function.

        Returns
        -------
        self : object
            Returns self.
        """
        X, y = validate_data(self, X, y)
        y_type = type_of_target(y, input_name='y', raise_unknown=True)
        if y_type != 'binary':
            raise ValueError(
                f'Unknown label type: Only binary classification is supported. The type of the target is {y_type}.'
            )
        self.classes_ = np.unique(y)
        if len(self.classes_) == 1:
            raise ValueError("Classifier can't train when only one class is present.")
        y = np.where(y == self.classes_[1], 1, 0)

        if isinstance(self.loss, Metric):
            fp_cost, fn_cost, tp_cost, tn_cost = self.loss._evaluate_costs(**loss_params)
        else:
            tp_cost, tn_cost, fn_cost, fp_cost = self._check_costs(
                tp_cost=tp_cost,
                tn_cost=tn_cost,
                fn_cost=fn_cost,
                fp_cost=fp_cost,
            )

        n_samples = X.shape[0]
        for name, cost in zip(
            ['tp_cost', 'tn_cost', 'fn_cost', 'fp_cost'], [tp_cost, tn_cost, fn_cost, fp_cost], strict=True
        ):
            if isinstance(cost, np.ndarray) and cost.shape[0] != n_samples:
                raise ValueError(f'{name} has shape {cost.shape}, but should have shape ({n_samples},)')

        if self.criterion == 'cost':
            self.criterion_ = CostImpurity(
                n_outputs=1,
                n_classes=np.array([2], dtype=np.intp),
            )
        elif self.criterion == 'gini':
            self.criterion_ = GiniCostImpurity(
                n_outputs=1,
                n_classes=np.array([2], dtype=np.intp),
            )
        elif self.criterion in {'entropy', 'log_loss'}:
            self.criterion_ = EntropyCostImpurity(
                n_outputs=1,
                n_classes=np.array([2], dtype=np.intp),
            )
        else:
            raise ValueError(f'Unknown criterion: {self.criterion}')

        self.criterion_.set_costs(
            tp_cost=tp_cost if not isinstance(tp_cost, np.ndarray) else 0.0,
            tn_cost=tn_cost if not isinstance(tn_cost, np.ndarray) else 0.0,
            fp_cost=fp_cost if not isinstance(fp_cost, np.ndarray) else 0.0,
            fn_cost=fn_cost if not isinstance(fn_cost, np.ndarray) else 0.0,
        )
        self.criterion_.set_array_costs(
            tp_cost=tp_cost.reshape(-1).astype(np.float64)
            if isinstance(tp_cost, np.ndarray)
            else np.array([], dtype=np.float64),
            tn_cost=tn_cost.reshape(-1).astype(np.float64)
            if isinstance(tn_cost, np.ndarray)
            else np.array([], dtype=np.float64),
            fp_cost=fp_cost.reshape(-1).astype(np.float64)
            if isinstance(fp_cost, np.ndarray)
            else np.array([], dtype=np.float64),
            fn_cost=fn_cost.reshape(-1).astype(np.float64)
            if isinstance(fn_cost, np.ndarray)
            else np.array([], dtype=np.float64),
            n_samples=n_samples,
        )

        self.estimator_ = RandomForestClassifier(
            n_estimators=self.n_estimators,
            criterion=self.criterion_,
            max_depth=self.max_depth,
            min_samples_split=self.min_samples_split,
            min_samples_leaf=self.min_samples_leaf,
            min_weight_fraction_leaf=self.min_weight_fraction_leaf,
            max_features=self.max_features,
            max_leaf_nodes=self.max_leaf_nodes,
            min_impurity_decrease=self.min_impurity_decrease,
            bootstrap=self.bootstrap,
            oob_score=self.oob_score,
            n_jobs=self.n_jobs,
            random_state=self.random_state,
            verbose=self.verbose,
            warm_start=self.warm_start,
            class_weight=self.class_weight,
            ccp_alpha=self.ccp_alpha,
            max_samples=self.max_samples,
            monotonic_cst=self.monotonic_cst,
        )
        self.estimator_.fit(X, y)

        if self.combination == 'weighted_voting':
            if not self.bootstrap:
                raise ValueError('Weighted voting is only available when bootstrap=True.')
            if self.loss is None:
                self.estimator_weights_ = self._get_oob_weights(
                    X,
                    y,
                    tp_cost=tp_cost,
                    tn_cost=tn_cost,
                    fn_cost=fn_cost,
                    fp_cost=fp_cost,
                    check_input=False,
                )
            else:
                self.estimator_weights_ = self._get_oob_weights(X, y, **loss_params)
        return self

    def predict(self, X: FloatArrayLike) -> IntNDArray:
        """
        Predict class of X.

        The predicted class for each sample in X is returned.

        Parameters
        ----------
        X : array-like of shape = [n_samples, n_features]
            The input samples.

        Returns
        -------
        y : array of shape = [n_samples]
            The predicted classes,
        """
        check_is_fitted(self)
        y_proba = self.predict_proba(X)
        y_pred: IntNDArray = self.classes_.take(np.argmax(y_proba, axis=1), axis=0)
        return y_pred

    def predict_proba(self, X: FloatArrayLike) -> FloatNDArray:
        """
        Predict class probabilities of the input samples X.

        Parameters
        ----------
        X : array-like of shape = [n_samples, n_features]
            The input samples.

        Returns
        -------
        prob : array of shape = [n_samples, 2]
            The class probabilities of the input samples.
        """
        check_is_fitted(self)
        X: FloatNDArray = validate_data(self, X, reset=False)

        if self.combination == 'weighted_voting':
            y_proba: FloatNDArray = self._predict_weighted_proba(X)
        else:
            y_proba = self.estimator_.predict_proba(X)

        return y_proba

    def predict_log_proba(self, X: FloatArrayLike) -> FloatNDArray:
        """
        Predict class log-probabilities for X.

        The predicted class log-probabilities of an input sample is computed as
        the log of the mean predicted class probabilities of the trees in the forest.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The input samples. Internally, its dtype will be converted to
            ``dtype=np.float32``. If a sparse matrix is provided, it will be
            converted into a sparse ``csr_matrix``.

        Returns
        -------
        p : ndarray of shape (n_samples, n_classes), or a list of such arrays
            The class probabilities of the input samples. The order of the
            classes corresponds to that in the attribute :term:`classes_`.
        """
        check_is_fitted(self)
        y_proba = self.predict_proba(X)
        return np.log(y_proba)

    def apply(self, X: FloatArrayLike) -> IntNDArray:
        """
        Apply trees in the forest to X, return leaf indices.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The input samples. Internally, its dtype will be converted to
            ``dtype=np.float32``. If a sparse matrix is provided, it will be
            converted into a sparse ``csr_matrix``.

        Returns
        -------
        X_leaves : ndarray of shape (n_samples, n_estimators)
            For each datapoint x in X and for each tree in the forest,
            return the index of the leaf x ends up in.
        """
        check_is_fitted(self)
        X_leaves: IntNDArray = self.estimator_.apply(X)
        return X_leaves

    def decision_path(self, X: FloatArrayLike) -> tuple[csr_matrix, IntNDArray]:
        """
        Return the decision path in the forest.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The input samples. Internally, its dtype will be converted to
            ``dtype=np.float32``. If a sparse matrix is provided, it will be
            converted into a sparse ``csr_matrix``.

        Returns
        -------
        indicator : sparse matrix of shape (n_samples, n_nodes)
            Return a node indicator matrix where non zero elements indicates
            that the samples goes through the nodes. The matrix is of CSR
            format.

        n_nodes_ptr : ndarray of shape (n_estimators + 1,)
            The columns from indicator[n_nodes_ptr[i]:n_nodes_ptr[i+1]]
            gives the indicator value for the i-th estimator.
        """
        check_is_fitted(self)
        indicator: csr_matrix
        n_nodes_ptr: IntNDArray
        indicator, n_nodes_ptr = self.estimator_.decision_path(X)
        return indicator, n_nodes_ptr

    def _get_oob_weights(self, X: FloatNDArray, y: IntNDArray, **kwargs: Any) -> FloatNDArray:
        # Prediction requires X to be in CSR format
        if issparse(X):
            X = X.tocsr()  # type: ignore[attr-defined]
        X = X.astype(np.float32)

        n_samples = y.shape[0]
        estimator_weights = np.zeros(self.n_estimators, dtype=np.float64)

        if self.max_samples is None:
            n_samples_bootstrap = n_samples

        if isinstance(self.max_samples, Integral):
            if self.max_samples > n_samples:
                msg = '`max_samples` must be <= n_samples={} but got value {}'
                raise ValueError(msg.format(n_samples, self.max_samples))
            n_samples_bootstrap = self.max_samples

        if isinstance(self.max_samples, Real):
            n_samples_bootstrap = max(round(n_samples * self.max_samples), 1)

        weight_fn = self.loss if isinstance(self.loss, Metric) else expected_cost_loss

        for i, estimator in enumerate(self.estimators_):
            unsampled_indices = _generate_unsampled_indices(
                estimator.random_state,
                n_samples,
                n_samples_bootstrap,
            )

            y_pred = self.estimator_._get_oob_predictions(estimator, X[unsampled_indices, :])
            estimator_weights[i] += weight_fn(y[unsampled_indices], y_pred[:, 1, 0], **kwargs)

        estimator_weights /= estimator_weights.sum()

        return estimator_weights

    def _predict_weighted_proba(self, X: FloatArrayLike) -> FloatNDArray:
        X: FloatNDArray = self.estimator_._validate_X_predict(X)

        # Assign chunk of trees to jobs
        n_jobs, _, _ = _partition_estimators(self.n_estimators, self.n_jobs)

        # avoid storing the output of every estimator by summing them here
        all_proba = np.zeros((X.shape[0], self.n_classes_), dtype=np.float64)  # type: ignore[arg-type]
        lock = threading.Lock()
        Parallel(n_jobs=n_jobs, verbose=self.verbose, require='sharedmem')(
            delayed(_accumulate_weighted_prediction)(e.predict_proba, X, all_proba, weight, lock)
            for e, weight in zip(self.estimators_, self.estimator_weights_, strict=True)
        )

        return all_proba


def _generate_unsampled_indices(random_state: int, n_samples: int, n_samples_bootstrap: int) -> IntNDArray:
    """Private function used to forest._set_oob_score function."""
    sample_indices = _generate_sample_indices(random_state, n_samples, n_samples_bootstrap)
    sample_counts = np.bincount(sample_indices, minlength=n_samples)
    unsampled_mask = sample_counts == 0
    indices_range = np.arange(n_samples)
    unsampled_indices: IntNDArray = indices_range[unsampled_mask]

    return unsampled_indices


def _generate_sample_indices(random_state: int, n_samples: int, n_samples_bootstrap: int) -> IntNDArray:
    """Private function used to _parallel_build_trees function."""
    random_instance = check_random_state(random_state)
    sample_indices: IntNDArray = random_instance.randint(0, n_samples, n_samples_bootstrap, dtype=np.int32)

    return sample_indices


def _accumulate_weighted_prediction(
    predict: Callable[..., FloatNDArray],
    X: FloatArrayLike,
    out: FloatNDArray,
    weight: float,
    lock: threading.Lock,
) -> None:
    """Calculate the weighted prediction."""
    prediction = predict(X, check_input=False)
    with lock:
        out += prediction * weight  # type: ignore[misc]


class CSBaggingClassifier(CostSensitiveMixin, ClassifierMixin, BaseEstimator):
    """
    Cost-sensitive Bagging classifier.

    A Bagging classifier is an ensemble meta-estimator that fits base
    classifiers each on random subsets of the original dataset and then
    aggregates their individual predictions (either by voting or by averaging)
    to form a final prediction. Such a meta-estimator can typically be used as
    a way to reduce the variance of a black-box estimator (e.g., a decision
    tree), by introducing randomization into its construction procedure and
    then making an ensemble out of it.

    This algorithm encompasses several works from the literature. When random
    subsets of the dataset are drawn as random subsets of the samples, then
    this algorithm is known as Pasting [1]_. If samples are drawn with
    replacement, then the method is known as Bagging [2]_. When random subsets
    of the dataset are drawn as random subsets of the features, then the method
    is known as Random Subspaces [3]_. Finally, when base estimators are built
    on subsets of both samples and features, then the method is known as
    Random Patches [4]_.

    Parameters
    ----------
    estimator : object, default=None
        The base estimator to fit on random subsets of the dataset.
        If None, then the base estimator is a :class:`~empulse.models.CSTreeClassifier`.

    n_estimators : int, default=10
        The number of base estimators in the ensemble.

    tp_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of true positives. If ``float``, then all true positives have the same cost.
        If array-like, then it is the cost of each true positive classification.
        Is overwritten if another `tp_cost` is passed to the ``fit`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit`` method.

    fp_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of false positives. If ``float``, then all false positives have the same cost.
        If array-like, then it is the cost of each false positive classification.
        Is overwritten if another `fp_cost` is passed to the ``fit`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit`` method.

    tn_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of true negatives. If ``float``, then all true negatives have the same cost.
        If array-like, then it is the cost of each true negative classification.
        Is overwritten if another `tn_cost` is passed to the ``fit`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit`` method.

    fn_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of false negatives. If ``float``, then all false negatives have the same cost.
        If array-like, then it is the cost of each false negative classification.
        Is overwritten if another `fn_cost` is passed to the ``fit`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit`` method.

    loss : Metric, default=None
        The loss function to use in order to evaluate the costs.
        If ``None``, then the costs provided to the constructor or to the ``fit``
        method are used directly.
        If a :class:``~empulse.metrics.Metric`` is provided, then the costs are computed using the
        metric, and any costs provided to the
        constructor or to the ``fit`` method are ignored.

    max_samples : int or float, default=1.0
        The number of samples to draw from X to train each base estimator (with
        replacement by default, see ``bootstrap`` for more details).

        - If int, then draw `max_samples` samples.
        - If float, then draw `max_samples * X.shape[0]` samples.

    max_features : int or float, default=1.0
        The number of features to draw from X to train each base estimator (
        without replacement by default, see ``bootstrap_features`` for more
        details).

        - If int, then draw ``max_features`` features.
        - If float, then draw ``max(1, int(max_features * n_features_in_))`` features.

    bootstrap : bool, default=True
        Whether samples are drawn with replacement. If False, sampling
        without replacement is performed.

    bootstrap_features : bool, default=False
        Whether features are drawn with replacement.

    oob_score : bool, default=False
        Whether to use out-of-bag samples to estimate
        the generalization error. Only available if bootstrap=True.

    warm_start : bool, default=False
        When set to True, reuse the solution of the previous call to fit
        and add more estimators to the ensemble, otherwise, just fit
        a whole new ensemble. See :term:`the Glossary <warm_start>`.

    n_jobs : int, default=None
        The number of jobs to run in parallel for both :meth:`fit` and
        :meth:`predict`. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all
        processors. See :term:`Glossary <n_jobs>` for more details.

    random_state : int, RandomState instance or None, default=None
        Controls the random resampling of the original dataset
        (sample wise and feature wise).
        If the base estimator accepts a `random_state` attribute, a different
        seed is generated for each instance in the ensemble.
        Pass an int for reproducible output across multiple function calls.
        See :term:`Glossary <random_state>`.

    verbose : int, default=0
        Controls the verbosity when fitting and predicting.

    Attributes
    ----------
    estimator_: estimator
        The base estimator from which the ensemble is grown.

    estimators_: list of estimators
        The collection of fitted base estimators.

    estimators_samples_: list of arrays
        The subset of drawn samples (i.e., the in-bag samples) for each base
        estimator.

    estimators_features_: list of arrays
        The subset of drawn features for each base estimator.

    References
    ----------

    .. [1] L. Breiman, "Pasting small votes for classification in large
           databases and on-line", Machine Learning, 36(1), 85-103, 1999.

    .. [2] L. Breiman, "Bagging predictors", Machine Learning, 24(2), 123-140,
           1996.

    .. [3] T. Ho, "The random subspace method for constructing decision
           forests", Pattern Analysis and Machine Intelligence, 20(8), 832-844,
           1998.

    .. [4] G. Louppe and P. Geurts, "Ensembles on Random Patches", Machine
           Learning and Knowledge Discovery in Databases, 346-361, 2012.

    .. [5] Correa Bahnsen, A., Aouada, D., & Ottersten, B.
           `"Ensemble of Example-Dependent Cost-Sensitive Decision Trees" <http://arxiv.org/abs/1505.04637>`__,
           2015, http://arxiv.org/abs/1505.04637.
    """

    _parameter_constraints: ClassVar[ParameterConstraint] = {
        'tp_cost': ['array-like', Real],
        'tn_cost': ['array-like', Real],
        'fn_cost': ['array-like', Real],
        'fp_cost': ['array-like', Real],
        'loss': [Metric, None],
        'combination': [
            StrOptions({'majority_voting', 'weighted_voting'}),
        ],
        **BaggingClassifier._parameter_constraints,
    }

    @property
    def estimators_(self) -> list[DecisionTreeClassifier]:
        """The collection of fitted sub-estimators."""
        check_is_fitted(self)
        estimators: list[DecisionTreeClassifier] = self.estimator_.estimators_
        return estimators

    @property
    def n_classes_(self) -> IntNDArray:
        """The number of classes seen during :term:`fit`."""
        check_is_fitted(self)
        n_classes: int | list[int] = self.estimator_.n_classes_
        n_classes_ = np.array(n_classes)
        return n_classes_

    @property
    def oob_score_(self) -> float:
        """Score of the training dataset obtained using an out-of-bag estimate."""
        check_is_fitted(self)
        oob_score: float = self.estimator_.oob_score_
        return oob_score

    @property
    def oob_decision_function_(self) -> FloatNDArray:
        """Decision function computed with out-of-bag estimate on the training set."""
        check_is_fitted(self)
        oob_decision_function: FloatNDArray = self.estimator_.oob_decision_function_
        return oob_decision_function

    @property
    def estimators_samples_(self) -> list[IntNDArray]:
        """The subset of drawn samples (i.e., the in-bag samples) for each base estimator."""
        check_is_fitted(self)
        estimators_samples: list[IntNDArray] = self.estimator_.estimators_samples_
        return estimators_samples

    @property
    def estimators_features_(self) -> list[IntNDArray]:
        """The subset of drawn features (i.e., the in-bag samples) for each base estimator."""
        check_is_fitted(self)
        estimators_samples: list[IntNDArray] = self.estimator_.estimators_features_
        return estimators_samples

    def __init__(
        self,
        estimator: Any = None,
        n_estimators: int = 10,
        *,
        tp_cost: FloatArrayLike | float = 0.0,
        tn_cost: FloatArrayLike | float = 0.0,
        fn_cost: FloatArrayLike | float = 0.0,
        fp_cost: FloatArrayLike | float = 0.0,
        loss: Metric | None = None,
        combination: Literal['majority_voting', 'weighted_voting'] = 'majority_voting',
        max_samples: float | int = 1.0,
        max_features: int | float = 1.0,
        bootstrap: bool = True,
        bootstrap_features: bool = False,
        oob_score: bool = False,
        warm_start: bool = False,
        n_jobs: int | None = None,
        random_state: int | np.random.RandomState | None = None,
        verbose: int = 0,
    ):
        self.estimator = estimator
        self.n_estimators = n_estimators
        self.tp_cost = tp_cost
        self.tn_cost = tn_cost
        self.fn_cost = fn_cost
        self.fp_cost = fp_cost
        self.loss = loss
        self.combination = combination
        self.max_samples = max_samples
        self.max_features = max_features
        self.bootstrap = bootstrap
        self.bootstrap_features = bootstrap_features
        self.oob_score = oob_score
        self.warm_start = warm_start
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.verbose = verbose

    def _more_tags(self) -> dict[str, bool]:
        return {
            'binary_only': True,
            'poor_score': True,
        }

    def __sklearn_tags__(self) -> Tags:
        tags = super().__sklearn_tags__()
        tags.classifier_tags.multi_class = False
        tags.classifier_tags.poor_score = True
        return tags

    def _get_metric_loss(self) -> Metric | None:
        """Get the metric loss function if available."""
        if isinstance(self.loss, Metric):
            return self.loss
        return None

    @_fit_context(prefer_skip_nested_validation=True)  # type: ignore[misc]
    def fit(
        self,
        X: FloatArrayLike,
        y: IntArrayLike,
        *,
        tp_cost: FloatArrayLike | float | Parameter = Parameter.UNCHANGED,
        tn_cost: FloatArrayLike | float | Parameter = Parameter.UNCHANGED,
        fn_cost: FloatArrayLike | float | Parameter = Parameter.UNCHANGED,
        fp_cost: FloatArrayLike | float | Parameter = Parameter.UNCHANGED,
        **loss_params: Any,
    ) -> Self:
        """
        Build a cost-sensitive bagging classifier from the training set.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples. Sparse matrices are accepted only if
            they are supported by the base estimator.

        y : array-like of shape (n_samples,)
            Ground truth (correct) labels.

        tp_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of true positives. If ``float``, then all true positives have the same cost.
            If array-like, then it is the cost of each true positive classification.

        fp_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of false positives. If ``float``, then all false positives have the same cost.
            If array-like, then it is the cost of each false positive classification.

        tn_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of true negatives. If ``float``, then all true negatives have the same cost.
            If array-like, then it is the cost of each true negative classification.

        fn_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of false negatives. If ``float``, then all false negatives have the same cost.
            If array-like, then it is the cost of each false negative classification.

        loss_params : dict
            Additional keyword arguments to pass to the loss function if using a custom loss function.

        Returns
        -------
        self : CSBaggingClassifier
            Returns self.
        """
        X, y = validate_data(self, X, y)
        y_type = type_of_target(y, input_name='y', raise_unknown=True)
        if y_type != 'binary':
            raise ValueError(
                f'Unknown label type: Only binary classification is supported. The type of the target is {y_type}.'
            )
        self.classes_ = np.unique(y)
        if len(self.classes_) == 1:
            raise ValueError("Classifier can't train when only one class is present.")
        y = np.where(y == self.classes_[1], 1, 0)

        if isinstance(self.loss, Metric):
            fp_cost, fn_cost, tp_cost, tn_cost = self.loss._evaluate_costs(**loss_params)
        else:
            tp_cost, tn_cost, fn_cost, fp_cost = self._check_costs(
                tp_cost=tp_cost,
                tn_cost=tn_cost,
                fn_cost=fn_cost,
                fp_cost=fp_cost,
            )

        n_samples = X.shape[0]
        for name, cost in zip(
            ['tp_cost', 'tn_cost', 'fn_cost', 'fp_cost'], [tp_cost, tn_cost, fn_cost, fp_cost], strict=True
        ):
            if isinstance(cost, np.ndarray) and cost.shape[0] != n_samples:
                raise ValueError(f'{name} has shape {cost.shape}, but should have shape ({n_samples},)')

        if self.estimator is None:
            criterion = CostImpurity(
                n_outputs=1,
                n_classes=np.array([2], dtype=np.intp),
            )
            criterion.set_costs(
                tp_cost=tp_cost if not isinstance(tp_cost, np.ndarray) else 0.0,
                tn_cost=tn_cost if not isinstance(tn_cost, np.ndarray) else 0.0,
                fp_cost=fp_cost if not isinstance(fp_cost, np.ndarray) else 0.0,
                fn_cost=fn_cost if not isinstance(fn_cost, np.ndarray) else 0.0,
            )
            criterion.set_array_costs(
                tp_cost=tp_cost.reshape(-1).astype(np.float64)
                if isinstance(tp_cost, np.ndarray)
                else np.array([], dtype=np.float64),
                tn_cost=tn_cost.reshape(-1).astype(np.float64)
                if isinstance(tn_cost, np.ndarray)
                else np.array([], dtype=np.float64),
                fp_cost=fp_cost.reshape(-1).astype(np.float64)
                if isinstance(fp_cost, np.ndarray)
                else np.array([], dtype=np.float64),
                fn_cost=fn_cost.reshape(-1).astype(np.float64)
                if isinstance(fn_cost, np.ndarray)
                else np.array([], dtype=np.float64),
                n_samples=n_samples,
            )
            self.base_estimator_ = CSTreeClassifier(criterion=criterion)
        else:
            self.base_estimator_ = clone(self.estimator)

        with config_context(enable_metadata_routing=True):
            self.estimator_ = BaggingClassifier(
                estimator=self.base_estimator_.set_fit_request(tp_cost=True, fp_cost=True, tn_cost=True, fn_cost=True),
                n_estimators=self.n_estimators,
                max_features=self.max_features,
                bootstrap=self.bootstrap,
                bootstrap_features=self.bootstrap_features,
                oob_score=self.oob_score,
                n_jobs=self.n_jobs,
                random_state=self.random_state,
                verbose=self.verbose,
                warm_start=self.warm_start,
                max_samples=self.max_samples,
            )
            self.estimator_.fit(X, y, tp_cost=tp_cost, tn_cost=tn_cost, fn_cost=fn_cost, fp_cost=fp_cost)

        if self.combination == 'weighted_voting':
            if not self.bootstrap:
                raise ValueError('Weighted voting is only available when bootstrap=True.')
            if self.loss is None:
                self.estimator_weights_ = self._get_oob_weights(
                    X,
                    y,
                    tp_cost=tp_cost,
                    tn_cost=tn_cost,
                    fn_cost=fn_cost,
                    fp_cost=fp_cost,
                    check_input=False,
                )
            else:
                self.estimator_weights_ = self._get_oob_weights(X, y, **loss_params)
        return self

    def predict(self, X: FloatArrayLike) -> IntNDArray:
        """Predict class for X.

        The predicted class of an input sample is computed as the class with
        the highest mean predicted probability. If base estimators do not
        implement a ``predict_proba`` method, then it resorts to voting.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples. Sparse matrices are accepted only if
            they are supported by the base estimator.

        Returns
        -------
        y : ndarray of shape (n_samples,)
            The predicted classes.
        """
        check_is_fitted(self)
        y_proba = self.predict_proba(X)
        y_pred: IntNDArray = self.classes_.take(np.argmax(y_proba, axis=1), axis=0)
        return y_pred

    def predict_proba(self, X: FloatArrayLike) -> FloatNDArray:
        """Predict class probabilities for X.

        The predicted class probabilities of an input sample is computed as
        the mean predicted class probabilities of the base estimators in the
        ensemble. If base estimators do not implement a ``predict_proba``
        method, then it resorts to voting and the predicted class probabilities
        of an input sample represents the proportion of estimators predicting
        each class.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples. Sparse matrices are accepted only if
            they are supported by the base estimator.

        Returns
        -------
        p : ndarray of shape (n_samples, n_classes)
            The class probabilities of the input samples. The order of the
            classes corresponds to that in the attribute :term:`classes_`.
        """
        check_is_fitted(self)
        X: FloatNDArray = validate_data(self, X, reset=False)

        if self.combination == 'weighted_voting':
            y_proba: FloatNDArray = self._predict_weighted_proba(X)
        else:
            y_proba = self.estimator_.predict_proba(X)

        return y_proba

    def predict_log_proba(self, X: FloatArrayLike) -> FloatNDArray:
        """
        Predict class log-probabilities for X.

        The predicted class log-probabilities of an input sample is computed as
        the log of the mean predicted class probabilities of the base
        estimators in the ensemble.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples. Sparse matrices are accepted only if
            they are supported by the base estimator.

        Returns
        -------
        p : ndarray of shape (n_samples, n_classes)
            The class log-probabilities of the input samples. The order of the
            classes corresponds to that in the attribute :term:`classes_`.
        """
        check_is_fitted(self)
        y_proba = self.predict_proba(X)
        return np.log(y_proba)

    @available_if(_estimator_has('decision_function', delegates=('base_estimator_', 'estimator')))
    def decision_function(self, X: FloatArrayLike) -> FloatNDArray:
        """
        Average of the decision functions of the base classifiers.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples. Sparse matrices are accepted only if
            they are supported by the base estimator.

        Returns
        -------
        score : ndarray of shape (n_samples, 1)
            The decision function of the input samples. The columns correspond
            to the classes in sorted order, as they appear in the attribute
            ``classes_``.
        """
        check_is_fitted(self)
        decisions: FloatNDArray = self.estimator_.decision_function(X)
        return decisions

    def _get_oob_weights(self, X: FloatNDArray, y: IntNDArray, **loss_params: Any) -> FloatNDArray:
        n_samples = y.shape[0]

        estimator_weights = np.zeros(self.n_estimators, dtype=np.float64)
        weight_fn = self.loss if self.loss is not None else expected_cost_loss

        for i, estimator, samples, features in zip(
            range(self.n_estimators), self.estimators_, self.estimators_samples_, self.estimators_features_, strict=True
        ):
            # Create mask for OOB samples
            mask = ~indices_to_mask(samples, n_samples)

            if hasattr(estimator, 'predict_proba'):
                y_pred = estimator.predict_proba((X[mask, :])[:, features])[:, 1]
            else:
                y_pred = estimator.predict((X[mask, :])[:, features])
            estimator_weights[i] = weight_fn(y[mask], y_pred, **loss_params)

        estimator_weights /= estimator_weights.sum()

        return estimator_weights

    def _predict_weighted_proba(self, X: FloatNDArray) -> FloatNDArray:
        X = validate_data(self, X, reset=False)

        # Assign chunk of trees to jobs
        n_jobs, _, _ = _partition_estimators(self.n_estimators, self.n_jobs)

        # avoid storing the output of every estimator by summing them here
        all_proba = np.zeros((X.shape[0], self.n_classes_), dtype=np.float64)
        lock = threading.Lock()
        Parallel(n_jobs=n_jobs, verbose=self.verbose, require='sharedmem')(
            delayed(_accumulate_weighted_prediction_non_tree)(e.predict_proba, X, all_proba, weight, lock)
            for e, weight in zip(self.estimators_, self.estimator_weights_, strict=True)
        )

        return all_proba


def _accumulate_weighted_prediction_non_tree(
    predict: Callable[[FloatNDArray], FloatNDArray],
    X: FloatNDArray,
    out: FloatNDArray,
    weight: float,
    lock: threading.Lock,
) -> None:
    """Calculate the weighted prediction."""
    prediction = predict(X)
    with lock:
        out += prediction * weight  # type: ignore[misc]



import sys
import warnings
from collections.abc import Callable
from typing import TYPE_CHECKING, Any, ClassVar, TypeVar

import numpy as np
from imblearn.base import BaseSampler
from numpy.typing import ArrayLike, NDArray
from sklearn.base import clone
from sklearn.utils import _safe_indexing
from sklearn.utils._param_validation import HasMethods, StrOptions

from .._types import FloatNDArray, IntNDArray, ParameterConstraint
from ..utils._sklearn_compat import ClassifierTags, Tags, type_of_target  # type: ignore
from ._strategies import Strategy

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if TYPE_CHECKING:  # pragma: no cover
    import pandas as pd

    _XT = TypeVar('_XT', NDArray[Any], pd.DataFrame, ArrayLike)
    _YT = TypeVar('_YT', NDArray[Any], pd.Series, ArrayLike)
else:
    _XT = TypeVar('_XT', NDArray[Any], ArrayLike)
    _YT = TypeVar('_YT', NDArray[Any], ArrayLike)

StrategyFn = Callable[[NDArray[Any], NDArray[Any]], int]


def _independent_pairs(y_true: ArrayLike, sensitive_feature: NDArray[Any]) -> int:
    """Determine promotion and demotion pairs so that y is statistically independent of sensitive feature."""
    sensitive_indices = np.where(sensitive_feature == 0)[0]
    not_sensititive_indices = np.where(sensitive_feature == 1)[0]
    n_sensitive = len(sensitive_indices)
    n_not_sensitive = len(not_sensititive_indices)
    n = n_sensitive + n_not_sensitive

    # no swapping needed if one of the groups is empty
    if n_sensitive == 0 or n_not_sensitive == 0:
        warnings.warn(
            'sensitive_feature only contains one class, no relabeling is performed.',
            UserWarning,
            stacklevel=2,
        )
        return 0

    pos_ratio_sensitive = np.sum(_safe_indexing(y_true, sensitive_indices)) / n_sensitive
    pos_ratio_not_sensitive = np.sum(_safe_indexing(y_true, not_sensititive_indices)) / n_not_sensitive

    discrimination = pos_ratio_not_sensitive - pos_ratio_sensitive

    # number of pairs to swap label
    return int(abs(round((discrimination * n_sensitive * n_not_sensitive) / n)))


class BiasRelabler(BaseSampler):  # type: ignore[misc]
    """
    Sampler which relabels instances to remove bias against a subgroup.

    Read more in the :ref:`User Guide <bias_mitigation>`.

    Parameters
    ----------
    estimator : Estimator instance
        Base estimator which is used to determine the number of promotion and demotion pairs.
    strategy : {'statistical parity', 'demographic parity'} or Callable, default='statistical parity'
        Determines how the group weights are computed.
        Group weights determine how many instances to relabel for each combination of target and sensitive_feature.

        - ``'statistical_parity'`` or ``'demographic parity'``: \
        probability of positive predictions are equal between subgroups of sensitive feature.

        - ``Callable``: function which computes the number of labels swaps based on the target and sensitive feature. \
        Callable accepts two arguments: \
        y_true and sensitive_feature and returns the number of pairs needed to be swapped.
    transform_feature : Optional[Callable[[numpy.ndarray], numpy.ndarray]], default=None
        Function which transforms sensitive feature before resampling the training data.
        The function takes in the sensitive feature in the form of a numpy.ndarray
        and outputs the transformed sensitive feature as a numpy.ndarray.
        This can be useful if you want to transform a continuous variable to a binary variable at fit time.

    Attributes
    ----------
    estimator_ : Estimator instance
        Fitted estimator.

    Examples
    --------

    .. code-block:: python

        import numpy as np
        from empulse.samplers import BiasRelabler
        from sklearn.datasets import make_classification
        from sklearn.linear_model import LogisticRegression

        X, y = make_classification()
        high_clv = np.random.randint(0, 2, y.shape)

        sampler = BiasRelabler(LogisticRegression())
        sampler.fit_resample(X, y, sensitive_feature=high_clv)

    Example with passing high-clv indicator through cross-validation:

    .. code-block:: python

        import numpy as np
        from empulse.samplers import BiasRelabler
        from imblearn.pipeline import Pipeline
        from sklearn import set_config
        from sklearn.datasets import make_classification
        from sklearn.linear_model import LogisticRegression
        from sklearn.model_selection import cross_val_score

        set_config(enable_metadata_routing=True)

        X, y = make_classification()
        high_clv = np.random.randint(0, 2, y.shape)

        pipeline = Pipeline([
            ('sampler', BiasRelabler(
                LogisticRegression()
            ).set_fit_resample_request(sensitive_feature=True)),
            ('model', LogisticRegression())
        ])

        cross_val_score(pipeline, X, y, params={'sensitive_feature': high_clv})

    Example with passing clv through a grid search and dynamically determining high_clv customer based on training data:

    .. code-block:: python

        import numpy as np
        from empulse.samplers import BiasRelabler
        from imblearn.pipeline import Pipeline
        from sklearn import set_config
        from sklearn.datasets import make_classification
        from sklearn.linear_model import LogisticRegression
        from sklearn.model_selection import GridSearchCV

        set_config(enable_metadata_routing=True)

        X, y = make_classification()
        clv = np.random.rand(y.size)

        def to_high_clv(clv: np.ndarray) -> np.ndarray:
            return (clv > np.median(clv)).astype(np.int8)

        pipeline = Pipeline([
            ('sampler', BiasRelabler(
                LogisticRegression(),
                transform_feature=to_high_clv
            ).set_fit_resample_request(sensitive_feature=True)),
            ('model', LogisticRegression())
        ])
        param_grid = {'model__C': np.logspace(-5, 2, 10)}

        grid_search = GridSearchCV(pipeline, param_grid=param_grid)
        grid_search.fit(X, y, sensitive_feature=clv)

    References
    ----------

    .. [1] Rahman, S., Janssens, B., & Bogaert, M. (2025).
           Profit-driven pre-processing in B2B customer churn modeling using fairness techniques.
           Journal of Business Research, 189, 115159. doi:10.1016/j.jbusres.2024.115159
    """

    _estimator_type: ClassVar[str] = 'sampler'
    _sampling_type: ClassVar[str] = 'bypass'
    _parameter_constraints: ClassVar[ParameterConstraint] = {
        'estimator': [HasMethods(['fit', 'predict_proba'])],
        'strategy': [StrOptions({'statistical parity', 'demographic parity'}), callable],
        'transform_feature': [callable, None],
    }
    _strategy_mapping: ClassVar[dict[Strategy, StrategyFn]] = {
        'statistical parity': _independent_pairs,
        'demographic parity': _independent_pairs,
    }

    if TYPE_CHECKING:  # pragma: no cover
        # BaseEstimator should dynamically generate the method signature at runtime
        def set_fit_resample_request(self, sensitive_feature: bool = False) -> Self:  # noqa: D102
            pass

    def __init__(
        self,
        estimator: Any,
        *,
        strategy: StrategyFn | Strategy = 'statistical parity',
        transform_feature: Callable[[NDArray[Any]], IntNDArray] | None = None,
    ):
        super().__init__()
        self.estimator = estimator
        self.transform_feature = transform_feature
        self.strategy = strategy

    def _more_tags(self) -> dict[str, bool]:
        return {
            'binary_only': True,
            'poor_score': True,
        }

    def __sklearn_tags__(self) -> Tags:
        tags = super().__sklearn_tags__()
        tags.classifier_tags = ClassifierTags(multi_class=False)
        return tags

    def fit_resample(self, X: _XT, y: _YT, *, sensitive_feature: ArrayLike | None = None) -> tuple[_XT, _YT]:
        """
        Fit the estimator and relabel the data according to the strategy.

        Parameters
        ----------
        X : 2D array-like, shape=(n_samples, n_features)
        y : 1D array-like, shape=(n_samples,)
        sensitive_feature : 1D array-like, shape=(n_samples,)
            Sensitive feature used to determine the number of promotion and demotion pairs.

        Returns
        -------
        X : 2D array-like, shape=(n_samples, n_features)
            Original training data.
        y : np.ndarray
            Relabeled target values.
        """
        X, y = super().fit_resample(X, y, sensitive_feature=sensitive_feature)
        X: _XT
        y: _YT
        return X, y

    def _fit_resample(
        self, X: NDArray[Any], y: NDArray[Any], *, sensitive_feature: ArrayLike | None = None
    ) -> tuple[NDArray[Any], NDArray[Any]]:
        """
        Fit the estimator and relabel the data according to the strategy.

        Parameters
        ----------
        X : 2D array-like, shape=(n_samples, n_features)
        y : 1D array-like, shape=(n_samples,)
        sensitive_feature : 1D array-like, shape=(n_samples,)
            Sensitive feature used to determine the number of promotion and demotion pairs.

        Returns
        -------
        X : 2D array-like, shape=(n_samples, n_features)
            Original training data.
        y : np.ndarray
            Relabeled target values.
        """
        sensitive_feature = np.asarray(sensitive_feature)
        y_type = type_of_target(y, input_name='y', raise_unknown=True)
        if y_type != 'binary':
            raise ValueError(f'Only binary classification is supported. The type of the target is {y_type}.')
        self.classes_: NDArray[np.int64] = np.unique(y)
        if len(self.classes_) == 1:
            return X, y
        y_binarized = np.where(y == self.classes_[1], 1, 0)

        if self.transform_feature is not None:
            sensitive_feature = self.transform_feature(sensitive_feature)

        self.estimator_ = clone(self.estimator)
        self.estimator_.fit(X, y)
        y_pred = self.estimator_.predict_proba(X)[:, 1]

        strategy = self._strategy_mapping[self.strategy] if isinstance(self.strategy, str) else self.strategy
        n_pairs = strategy(y_binarized, sensitive_feature)
        if n_pairs <= 0:
            return X, np.asarray(y)

        sensitive_indices = np.where(sensitive_feature == 0)[0]
        non_sensitive = np.where(sensitive_feature == 1)[0]
        probas_non_sensitive = y_pred[non_sensitive]
        probas_sensitive = y_pred[sensitive_indices]

        demotion_candidates = _get_demotion_candidates(probas_non_sensitive, _safe_indexing(y, non_sensitive), n_pairs)
        promotion_candidates = _get_promotion_candidates(
            probas_sensitive, _safe_indexing(y, sensitive_indices), n_pairs
        )

        # map promotion and demotion candidates to original indices
        indices = np.arange(len(y))
        demotion_candidates = indices[non_sensitive][demotion_candidates]
        promotion_candidates = indices[sensitive_indices][promotion_candidates]

        # relabel the data
        if hasattr(y, 'copy'):
            relabled_y = y.copy()
        elif hasattr(y, 'clone'):
            relabled_y = y.clone()
        else:
            relabled_y = np.copy(y)

        if hasattr(relabled_y, 'loc'):
            relabled_y.loc[demotion_candidates] = 0
            relabled_y.loc[promotion_candidates] = 1
        else:
            relabled_y[demotion_candidates] = 0
            relabled_y[promotion_candidates] = 1

        return X, relabled_y


def _get_demotion_candidates(y_pred: FloatNDArray, y_true: FloatNDArray, n_pairs: int) -> FloatNDArray:
    """Return the n_pairs instances with the lowest probability of being positive class label."""
    positive_indices = np.where(y_true == 1)[0]
    positive_predictions = y_pred[positive_indices]
    demotion_candidates: FloatNDArray = positive_indices[np.argsort(positive_predictions)[:n_pairs]]
    return demotion_candidates


def _get_promotion_candidates(y_pred: FloatNDArray, y_true: FloatNDArray, n_pairs: int) -> FloatNDArray:
    """Return the n_pairs instances with the lowest probability of being negative class label."""
    negative_indices = np.where(y_true == 0)[0]
    negative_predictions = y_pred[negative_indices]
    promotion_candidates: FloatNDArray = negative_indices[np.argsort(negative_predictions)[-n_pairs:]]
    return promotion_candidates



from .cost_matrix import CostMatrix
from .cost_metric import Cost
from .metric import Metric
from .metric_strategies import MaxProfit
from .savings_metric import Savings


def make_generic_cost_matrix() -> CostMatrix:
    """Create a generic cost matrix."""
    return CostMatrix().add_tp_cost('tp_cost').add_tn_cost('tn_cost').add_fp_cost('fp_cost').add_fn_cost('fn_cost')


def make_generic_cost_metric() -> Metric:
    """Create a generic cost metric."""
    return Metric(cost_matrix=make_generic_cost_matrix(), strategy=Cost())


def make_generic_savings_metric() -> Metric:
    """Create a generic savings metric."""
    return Metric(cost_matrix=make_generic_cost_matrix(), strategy=Savings())


def make_generic_max_profit_metric() -> Metric:
    """Create a generic maximum profit metric."""
    return Metric(cost_matrix=make_generic_cost_matrix(), strategy=MaxProfit())



from collections.abc import Callable, Sequence
from functools import partial, update_wrapper
from typing import TYPE_CHECKING, Literal, TypeVar, overload

import numpy as np
from scipy.special import expit

from ..._types import FloatArrayLike, FloatNDArray

if TYPE_CHECKING:  # pragma: no cover
    try:
        from lightgbm import Dataset
        from xgboost import DMatrix

        Matrix = TypeVar('Matrix', bound=FloatNDArray | DMatrix | Dataset)
    except ImportError:
        Matrix = TypeVar('Matrix', bound=FloatNDArray)  # type: ignore[misc]
else:
    Matrix = TypeVar('Matrix', bound=FloatNDArray)


from ._validation import _validate_input_deterministic


@overload
def make_objective_acquisition(
    model: Literal['catboost'],
    *,
    contribution: float = 7_000,
    contact_cost: float = 50,
    sales_cost: float = 500,
    direct_selling: float = 1,
    commission: float = 0.1,
) -> tuple['AECObjectiveAcquisition', 'AECMetricAcquisition']: ...


@overload
def make_objective_acquisition(
    model: Literal['xgboost', 'lightgbm'],
    *,
    contribution: float = 7_000,
    contact_cost: float = 50,
    sales_cost: float = 500,
    direct_selling: float = 1,
    commission: float = 0.1,
) -> Callable[[FloatNDArray, FloatNDArray], tuple[FloatNDArray, FloatNDArray]]: ...


def make_objective_acquisition(
    model: Literal['xgboost', 'lightgbm', 'catboost'],
    *,
    contribution: float = 7_000,
    contact_cost: float = 50,
    sales_cost: float = 500,
    direct_selling: float = 1,
    commission: float = 0.1,
) -> (
    tuple['AECObjectiveAcquisition', 'AECMetricAcquisition']
    | Callable[[FloatNDArray, FloatNDArray], tuple[FloatNDArray, FloatNDArray]]
):
    """
    Create an objective function for the Expected Cost measure for customer acquisition.

    The objective function presumes a situation where leads are targeted either directly or indirectly.
    Directly targeted leads are contacted and handled by the internal sales team.
    Indirectly targeted leads are contacted and then referred to intermediaries,
    which receive a commission.
    The company gains a contribution from a successful acquisition.

    Read more in the :ref:`User Guide <cost_functions>`.

    Parameters
    ----------
    model : {'xgboost', 'lightgbm', 'catboost'}
        The model for which the objective function is created.

        - 'xgboost' : :class:`xgboost:xgboost.XGBClassifier`
        - 'lightgbm' : :class:`lightgbm:lightgbm.LGBMClassifier`
        - 'catboost' : :class:`catboost.CatBoostClassifier`

    contribution : float, default=7000
        Average contribution of a new customer (``contribution ≥ 0``).

    sales_cost : float, default=500
        Average sale conversion cost of targeted leads handled by the company (``sales_cost ≥ 0``).

    contact_cost : float, default=50
        Average contact cost of targeted leads (``contact_cost ≥ 0``).

    direct_selling : float, default=1
        Fraction of leads sold to directly (``0 ≤ direct_selling ≤ 1``).
        ``direct_selling = 0`` for indirect channel.
        ``direct_selling = 1`` for direct channel.

    commission : float, default=0.1
        Fraction of contribution paid to the intermediaries (``0 ≤ commission ≤ 1``).

        .. note::
            The commission is only relevant when there is an indirect channel (``direct_selling < 1``).

    Returns
    -------
    objective : Callable
        A custom objective function for XGBoost.


    Examples
    --------

    .. code-block::  python

        from xgboost import XGBClassifier
        from empulse.metrics import make_objective_acquisition

        objective = make_objective_acquisition(model='xgboost')
        clf = XGBClassifier(objective=objective, n_estimators=100, max_depth=3)

    References
    ----------
    .. [1] Janssens, B., Bogaert, M., Bagué, A., & Van den Poel, D. (2022).
        B2Boost: Instance-dependent profit-driven modelling of B2B churn.
        Annals of Operations Research, 1-27.
    """
    if model == 'xgboost':
        objective: Callable[[FloatNDArray, FloatNDArray], tuple[FloatNDArray, FloatNDArray]] = partial(
            _objective,
            contribution=contribution,
            contact_cost=contact_cost,
            sales_cost=sales_cost,
            direct_selling=direct_selling,
            commission=commission,
        )
        update_wrapper(objective, _objective)
    elif model == 'lightgbm':

        def objective(y_true: FloatNDArray, y_score: FloatNDArray) -> tuple[FloatNDArray, FloatNDArray]:
            """
            Create an objective function for the churn AEC measure.

            Parameters
            ----------
            y_true : np.ndarray
                Ground truth labels.
            y_score : np.ndarray
                Predicted values.

            Returns
            -------
            gradient  : np.ndarray
                Gradient of the objective function.

            hessian : np.ndarray
                Hessian of the objective function.
            """
            return _objective(
                y_true,
                y_score,
                contribution=contribution,
                contact_cost=contact_cost,
                sales_cost=sales_cost,
                direct_selling=direct_selling,
                commission=commission,
            )

    elif model == 'catboost':
        return (
            AECObjectiveAcquisition(
                contribution=contribution,
                contact_cost=contact_cost,
                sales_cost=sales_cost,
                direct_selling=direct_selling,
                commission=commission,
            ),
            AECMetricAcquisition(
                contribution=contribution,
                contact_cost=contact_cost,
                sales_cost=sales_cost,
                direct_selling=direct_selling,
                commission=commission,
            ),
        )
    else:
        raise ValueError(f"Expected model to be 'xgboost' or 'lightgbm', got {model} instead.")
    return objective


def _objective(
    y_true: FloatNDArray,
    y_score: FloatNDArray,
    contribution: float = 7_000,
    contact_cost: float = 50,
    sales_cost: float = 500,
    direct_selling: float = 1,
    commission: float = 0.1,
) -> tuple[FloatNDArray, FloatNDArray]:
    """
    Create an objective function for `XGBoostClassifier` for customer acquisition.

    Parameters
    ----------
    y_true : np.ndarray
        Ground truth labels.
    y_score : np.ndarray
        Predicted values.

    Returns
    -------
    gradient  : np.ndarray
        Gradient of the objective function.

    hessian : np.ndarray
        Hessian of the objective function.
    """
    y_proba = expit(y_score)
    cost = (
        y_true
        * (
            direct_selling * (contact_cost + sales_cost - contribution)
            + (1 - direct_selling) * (contact_cost - (1 - commission) * contribution)
        )
        + (1 - y_true) * contact_cost
    )
    gradient = y_proba * (1 - y_proba) * cost
    hessian = np.abs((1 - 2 * y_proba) * gradient)
    return gradient, hessian


class AECObjectiveAcquisition:
    """AEC acquisition objective for catboost."""

    def __init__(
        self,
        contribution: float = 7_000,
        contact_cost: float = 50,
        sales_cost: float = 500,
        direct_selling: float = 1,
        commission: float = 0.1,
    ):
        self.contribution = contribution
        self.sales_cost = sales_cost
        self.contact_cost = contact_cost
        self.direct_selling = direct_selling
        self.commission = commission

    def calc_ders_range(
        self, predictions: Sequence[float], targets: FloatNDArray, weights: Sequence[float]
    ) -> list[tuple[float, float]]:
        """
        Compute first and second derivative of the loss function with respect to the predicted value for each object.

        Parameters
        ----------
        predictions : indexed container of floats
            Current predictions for each object.

        targets : indexed container of floats
            Target values you provided with the dataset.

        weights : float, optional (default=None)
            Instance weight.

        Returns
        -------
            der1 : list-like object of float
            der2 : list-like object of float

        """
        y_proba = expit(predictions)
        cost = (
            targets
            * (
                self.direct_selling * (self.contact_cost + self.sales_cost - self.contribution)
                + (1 - self.direct_selling) * (self.contact_cost - (1 - self.commission) * self.contribution)
            )
            + (1 - targets) * self.contact_cost
        )
        gradient = y_proba * (1 - y_proba) * cost
        hessian = np.abs((1 - 2 * y_proba) * gradient)
        return list(zip(-gradient, -hessian, strict=False))


class AECMetricAcquisition:
    """AEC acquisition metric for catboost."""

    def __init__(
        self,
        contribution: float = 7_000,
        contact_cost: float = 50,
        sales_cost: float = 500,
        direct_selling: float = 1,
        commission: float = 0.1,
    ):
        self.contribution = contribution
        self.sales_cost = sales_cost
        self.contact_cost = contact_cost
        self.direct_selling = direct_selling
        self.commission = commission

    def is_max_optimal(self) -> bool:
        """Return whether great values of metric are better."""
        return False

    def evaluate(
        self, predictions: Sequence[float], targets: Sequence[float], weights: Sequence[float]
    ) -> tuple[float, float]:
        """
        Evaluate metric value.

        Parameters
        ----------
        predictions : list of indexed containers (containers with only __len__ and __getitem__ defined) of float
            Vectors of approx labels.

        targets : one dimensional indexed container of float
            Vectors of true labels.

        weights : one dimensional indexed container of float, optional (default=None)
            Weight for each instance.

        Returns
        -------
            weighted error : float
            total weight : float

        """
        y_proba = expit(predictions)
        return expected_cost_loss_acquisition(
            targets,
            y_proba,
            contribution=self.contribution,
            contact_cost=self.contact_cost,
            sales_cost=self.sales_cost,
            direct_selling=self.direct_selling,
            commission=self.commission,
            normalize=True,
            check_input=False,
        ), 1

    def get_final_error(self, error: float, weight: float) -> float:
        """
        Return final value of metric based on error and weight.

        Parameters
        ----------
        error : float
            Sum of errors in all instances.

        weight : float
            Sum of weights of all instances.

        Returns
        -------
        metric value : float

        """
        return error


def expected_cost_loss_acquisition(
    y_true: FloatArrayLike,
    y_proba: FloatArrayLike,
    *,
    contribution: float = 7_000,
    contact_cost: float = 50,
    sales_cost: float = 500,
    direct_selling: float = 1,
    commission: float = 0.1,
    normalize: bool = False,
    check_input: bool = True,
) -> float:
    """
    Expected cost of a classifier for customer acquisition.

    The cost function presumes a situation where leads are targeted either directly or indirectly.
    Directly targeted leads are contacted and handled by the internal sales team.
    Indirectly targeted leads are contacted and then referred to intermediaries,
    which receive a commission.
    The company gains a contribution from a successful acquisition.

    Parameters
    ----------
    y_true : 1D array-like, shape=(n_samples,)
        Binary target values ('churn': 1, 'no churn': 0).

    y_proba : 1D array-like, shape=(n_samples,)
        Target probabilities, should lie between 0 and 1.

    contribution : float, default=7000
        Average contribution of a new customer (``contribution ≥ 0``).

    sales_cost : float, default=500
        Average sale conversion cost of targeted leads handled by the company (``sales_cost ≥ 0``).

    contact_cost : float, default=50
        Average contact cost of targeted leads (``contact_cost ≥ 0``).

    direct_selling : float, default=1
        Fraction of leads sold to directly (``0 ≤ direct_selling ≤ 1``).
        ``direct_selling = 0`` for indirect channel.
        ``direct_selling = 1`` for direct channel.

    commission : float, default=0.1
        Fraction of contribution paid to the intermediaries (``0 ≤ commission ≤ 1``).

        .. note::
            The commission is only relevant when there is an indirect channel (``direct_selling < 1``).

    normalize : bool, default=True
        Normalize the cost function by the number of samples.
        If ``True``, return the average expected cost for customer acquisition.

    check_input : bool, default=True
        Perform input validation.
        Turning off improves performance, useful when using this metric as a loss function.

    Returns
    -------
    empa_cost : float
        Instance-specific cost function according to the EMPA measure.


    """  # noqa: D401
    if check_input:
        y_true, y_proba = _validate_input_deterministic(
            y_true, y_proba, contribution, contact_cost, sales_cost, direct_selling, commission
        )
    else:
        y_true = np.asarray(y_true)
        y_proba = np.asarray(y_proba)

    costs = (
        y_true
        * y_proba
        * (
            direct_selling * (sales_cost + contact_cost - contribution)
            + (1 - direct_selling) * (contact_cost - (1 - commission) * contribution)
        )
        + (1 - y_true) * y_proba * contact_cost
    )
    if normalize:
        return float(np.mean(costs))
    return float(np.sum(costs))



import sys
import warnings
from typing import TYPE_CHECKING, Any, ClassVar, Literal

import numpy as np
from imblearn.base import BaseSampler
from numpy.typing import ArrayLike, NDArray
from sklearn.utils import check_random_state
from sklearn.utils._param_validation import Interval, Real, StrOptions

from .._common import Parameter
from .._types import FloatArrayLike, IntNDArray, ParameterConstraint
from ..utils._sklearn_compat import ClassifierTags, Tags  # type: ignore

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class CostSensitiveSampler(BaseSampler):  # type: ignore[misc]
    """
    Sampler which performs cost-proportionate resampling.

    This method adjusts the sampling probability of each sample based on the cost of misclassification.
    This is done either by rejection sampling [1]_ or oversampling [2]_.

    Read more in the :ref:`User Guide <cost_sampling>`.

    Parameters
    ----------
    method : {'rejection sampling', 'oversampling'}, default='rejection sampling'
        Method to perform the cost-proportionate sampling,
        either 'RejectionSampling' or 'OverSampling'.

    oversampling_norm: float, default=0.1
        Oversampling norm for the cost.
        The smaller the oversampling_norm, the more samples are generated.

    percentile_threshold: float, default=0.975
        Outlier adjustment for the cost.
        Costs are normalized and cost values above the percentile_threshold'th percentile are set to 1.

    random_state : int or :class:`numpy:numpy.random.RandomState`, optional
        Random number generator seed for reproducibility.

    fp_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of false positives. If ``float``, then all false positives have the same cost.
        If array-like, then it is the cost of each false positive classification.
        Is overwritten if another `fp_cost` is passed to the ``fit_resample`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit_resample`` method.

    fn_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of false negatives. If ``float``, then all false negatives have the same cost.
        If array-like, then it is the cost of each false negative classification.
        Is overwritten if another `fn_cost` is passed to the ``fit`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit_resample`` method.

    Attributes
    ----------
    sample_indices_ : numpy.ndarray
        Indices of the samples that were selected.

    References
    ----------
    .. [1] B. Zadrozny, J. Langford, N. Naoki, "Cost-sensitive learning by
           cost-proportionate example weighting", in Proceedings of the
           Third IEEE International Conference on Data Mining, 435-442, 2003.

    .. [2] C. Elkan, "The foundations of Cost-Sensitive Learning",
           in Seventeenth International Joint Conference on Artificial Intelligence,
           973-978, 2001.

    Notes
    -----
    code modified from `costcla.sampling.cost_sampling`.

    Examples
    --------

    .. code-block:: python

        import numpy as np
        from empulse.samplers import CostSensitiveSampler
        from sklearn.datasets import make_classification

        X, y = make_classification()
        fp_cost = np.ones_like(y) * 10
        fn_cost = np.ones_like(y)

        sampler = CostSensitiveSampler(method='oversampling', random_state=42)
        X_re, y_re = sampler.fit_resample(X, y, fp_cost=fp_cost, fn_cost=fn_cost)

    """

    _sampling_type: ClassVar[str] = 'bypass'
    _parameter_constraints: ClassVar[ParameterConstraint] = {
        'method': [StrOptions({'oversampling', 'rejection sampling'})],
        'oversampling_norm': [Interval(Real, 0, 1, closed='both')],
        'percentile_threshold': [Interval(Real, 0, 1, closed='both')],
        'random_state': ['random_state'],
        'fp_cost': [Real, 'array-like'],
        'fn_cost': [Real, 'array-like'],
    }

    if TYPE_CHECKING:  # pragma: no cover
        # BaseEstimator should dynamically generate the method signature at runtime
        def set_fit_resample_request(self, *, fp_cost: bool = False, fn_cost: bool = False) -> Self:  # noqa: D102
            pass

    def __init__(
        self,
        method: Literal['rejection sampling', 'oversampling'] = 'rejection sampling',
        *,
        oversampling_norm: float = 0.1,
        percentile_threshold: float = 0.975,
        random_state: int | np.random.RandomState | None = None,
        fp_cost: float | FloatArrayLike = 0.0,
        fn_cost: float | FloatArrayLike = 0.0,
    ):
        super().__init__()
        self.method = method
        self.oversampling_norm = oversampling_norm
        self.percentile_threshold = percentile_threshold
        self.random_state = random_state
        self.fp_cost = fp_cost
        self.fn_cost = fn_cost

    def _more_tags(self) -> dict[str, bool]:
        return {
            'binary_only': True,
            'poor_score': True,
        }

    def __sklearn_tags__(self) -> Tags:
        tags = super().__sklearn_tags__()
        tags.classifier_tags = ClassifierTags(multi_class=False)
        tags.sampler_tags.sample_indices = True
        return tags

    def fit_resample(
        self,
        X: ArrayLike,
        y: ArrayLike,
        *,
        fp_cost: float | ArrayLike | Parameter = Parameter.UNCHANGED,
        fn_cost: float | ArrayLike | Parameter = Parameter.UNCHANGED,
    ) -> tuple[NDArray[Any], NDArray[Any]]:
        """
        Resample the dataset.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)

        y : array-like of shape (n_samples,)

        fp_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of false positives. If ``float``, then all false positives have the same cost.
            If array-like, then it is the cost of each false positive classification.

        fn_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of false negatives. If ``float``, then all false negatives have the same cost.
            If array-like, then it is the cost of each false negative classification.

        Returns
        -------
        X_resampled : ndarray of shape (n_samples_new, n_features)
            The array containing the resampled data.

        y_resampled : ndarray of shape (n_samples_new,)
            The corresponding label of `X_resampled`.
        """
        X, y = super().fit_resample(X, y, fp_cost=fp_cost, fn_cost=fn_cost)
        X: NDArray[Any]
        y: NDArray[Any]
        return X, y

    def _fit_resample(
        self,
        X: NDArray[Any],
        y: IntNDArray,
        fp_cost: float | FloatArrayLike | Parameter = 0.0,
        fn_cost: float | FloatArrayLike | Parameter = 0.0,
    ) -> tuple[NDArray[Any], NDArray[Any]]:
        if fp_cost is Parameter.UNCHANGED:
            fp_cost = self.fp_cost
        if fn_cost is Parameter.UNCHANGED:
            fn_cost = self.fn_cost

        if (
            all(isinstance(cost, Real) for cost in (fp_cost, fn_cost))
            and sum(abs(cost) for cost in (fp_cost, fn_cost)) == 0.0  # type: ignore[misc, arg-type]
        ):
            warnings.warn(
                'All costs are zero. Setting fp_cost=1 and fn_cost=1. '
                f'To avoid this warning, set costs explicitly in the {self.__class__.__name__}.fit_resample() method.',
                UserWarning,
                stacklevel=2,
            )
            fp_cost = 1
            fn_cost = 1

        fp_cost = np.full_like(y, fp_cost) if isinstance(fp_cost, Real) else np.asarray(fp_cost)
        fn_cost = np.full_like(y, fn_cost) if isinstance(fn_cost, Real) else np.asarray(fn_cost)
        rng = check_random_state(self.random_state)

        misclassification_costs = fp_cost
        misclassification_costs[y == 1] = fn_cost[y == 1]

        normalized_costs = np.minimum(
            misclassification_costs / np.percentile(misclassification_costs, self.percentile_threshold * 100), 1
        )

        n_samples = X.shape[0]

        if self.method == 'rejection sampling':
            rejection_probability = rng.rand(n_samples)
            self.sample_indices_ = np.arange(len(y))[rejection_probability <= normalized_costs]
        elif self.method == 'oversampling':
            # repeat each sample based on the normalized costs
            sample_repeats = np.ceil(normalized_costs / self.oversampling_norm).astype(np.int64)
            self.sample_indices_ = np.repeat(np.arange(n_samples), sample_repeats)
        else:
            raise ValueError(f'Method not valid. Expected RejectionSampling or OverSampling, got {self.method}.')

        X_re = X[self.sample_indices_]
        y_re = y[self.sample_indices_]

        return X_re, y_re
